import unittest
from unittest import mock

import discord

from lgrez import config, bdd
from lgrez.features import IA
from test import mock_discord, mock_bdd


class TestIAFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.IA utility functions."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    async def test__build_sequence(self):
        """Unit tests for IA._build_sequence function."""
        # async def _build_sequence(ctx)
        _build_sequence = IA._build_sequence

        # message
        ctx = mock_discord.get_ctx(None)
        with mock_discord.interact(
            ("wait_for_react_clic", "mezzag"),
            ("wait_for_react_clic", False),
        ):
            seq = await _build_sequence(ctx)
        self.assertEqual(seq, "mezzag")
        ctx.assert_sent("Réaction du bot", "Puis")

        # command
        ctx = mock_discord.get_ctx(None)
        with mock_discord.interact(
            ("wait_for_react_clic", "!commandz"),
            ("wait_for_react_clic", False),
        ):
            seq = await _build_sequence(ctx)
        self.assertEqual(seq, f"{IA.MARK_CMD}commandz")
        ctx.assert_sent("Réaction du bot", "Puis")

        # command - other prefix
        ctx = mock_discord.get_ctx(None)
        config.bot.command_prefix = "PREFIXZ"
        with mock_discord.interact(
            ("wait_for_react_clic", "PREFIXZcommandz"),
            ("wait_for_react_clic", False),
        ):
            seq = await _build_sequence(ctx)
        self.assertEqual(seq, f"{IA.MARK_CMD}commandz")
        ctx.assert_sent("Réaction du bot", "Puis")
        config.bot.command_prefix = "!"

        # react
        ctx = mock_discord.get_ctx(None)
        reaczt = mock.NonCallableMock(discord.Emoji)
        reaczt.configure_mock(name="bon.")
        with mock_discord.interact(
            ("wait_for_react_clic", reaczt),
            ("wait_for_react_clic", False),
        ):
            seq = await _build_sequence(ctx)
        self.assertEqual(seq, f"{IA.MARK_REACT}bon.")
        ctx.assert_sent("Réaction du bot", "Puis")

        # several commands
        ctx = mock_discord.get_ctx(None)
        reaczt = mock.NonCallableMock(discord.Emoji)
        reaczt.configure_mock(name="bon.")
        with mock_discord.interact(
            ("wait_for_react_clic", "mezzage"),
            ("wait_for_react_clic", IA.MARK_THEN),
            ("wait_for_react_clic", reaczt),
            ("wait_for_react_clic", IA.MARK_OR),
            ("wait_for_react_clic", "oh"),
            ("wait_for_react_clic", IA.MARK_THEN),
            ("wait_for_react_clic", "!ouizz"),
            ("wait_for_react_clic", False),
        ):
            seq = await _build_sequence(ctx)
        self.assertEqual(seq,
            f"mezzage{IA.MARK_THEN}{IA.MARK_REACT}bon.{IA.MARK_OR}"
            f"oh{IA.MARK_THEN}{IA.MARK_CMD}ouizz"
        )
        ctx.assert_sent("Réaction du bot", "Puis", "Réaction du bot", "Puis",
                        "Réaction du bot", "Puis", "Réaction du bot", "Puis")


    def test_fetch_tenor(self):
        """Unit tests for IA.fetch_tenor function."""
        # def fetch_tenor(trigger)
        fetch_tenor = IA.fetch_tenor

        # request failed
        payload = mock.NonCallableMagicMock(__bool__=lambda self: False)
        with mock.patch("requests.get", return_value=payload):
            gif = fetch_tenor("bzzt")
        self.assertIs(None, gif)

        # no results
        payload = mock.NonCallableMagicMock(__bool__=lambda self: True,
                                            json=lambda: {"results": []})
        with mock.patch("requests.get", return_value=payload):
            gif = fetch_tenor("bzzt")
        self.assertIs(None, gif)

        # one result
        payload = mock.NonCallableMagicMock(
            __bool__=lambda self: True,
            json=lambda: {"results": [{"itemurl": "bzoot"}]}
        )
        with mock.patch("requests.get", return_value=payload):
            gif = fetch_tenor("bzzt")
        self.assertEqual("bzoot", gif)

        # more results
        payload = mock.NonCallableMagicMock(
            __bool__=lambda self: True,
            json=lambda: {"results": [{"itemurl": "bzoot"},
                                      {"itemurl": "notyou"},
                                      {"itemurl": "uneither"}]}
        )
        with mock.patch("requests.get", return_value=payload):
            gif = fetch_tenor("bzzt")
        self.assertEqual("bzoot", gif)


    async def test_trigger_at_mj(self):
        """Unit tests for IA.trigger_at_mj function."""
        # async def trigger_at_mj(message)
        trigger_at_mj = IA.trigger_at_mj

        # no mention
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "oo", role_mentions=[])
        rep = await trigger_at_mj(message)
        self.assertIs(rep, False)
        ctx.assert_sent()

        # other role mention
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "oo",
                                       role_mentions=[config.Role.joueur_mort])
        rep = await trigger_at_mj(message)
        self.assertIs(rep, False)
        ctx.assert_sent()

        # @MJ mention
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "oo",
                                       role_mentions=[config.Role.mj])
        rep = await trigger_at_mj(message)
        self.assertIs(rep, True)
        ctx.assert_sent("ils sont en route")


    @mock_bdd.patch_db      # Empty database for this method
    async def test_trigger_roles(self):
        """Unit tests for IA.trigger_roles function."""
        # async def trigger_roles(message, sensi=0.8)
        trigger_roles = IA.trigger_roles
        mock_bdd.add_campsroles(10, 10)

        # nothing
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "ooooo")
        rep = await trigger_roles(message)
        self.assertIs(rep, False)
        ctx.assert_sent()

        # one exact role found
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "Role3")
        role3 = bdd.Role.query.get("role3")
        role3.prefixe = "PREF@x"
        role3.description_courte = "D€$k00rt"
        role3.description_longue = "D€$k_longue"
        role3.camp = bdd.Camp.query.get("camp3")
        rep = await trigger_roles(message)
        self.assertIs(rep, True)
        ctx.assert_sent(["Role3", "PREF@x", "D€$k00rt", "D€$k_longue", "Camp3"])

        # high sensi
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "Rol")
        rep = await trigger_roles(message)
        self.assertIs(rep, False)
        ctx.assert_sent()

        # low sensi - more roles found
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "Rol")
        rep = await trigger_roles(message, sensi=0.5)
        self.assertIs(rep, True)
        ctx.assert_sent("Role")     # one selected


    @mock_bdd.patch_db      # Empty database for this method
    async def test_trigger_reactions(self):
        """Unit tests for IA.trigger_reactions function."""
        # async def trigger_reactions(message, chain=None, sensi=0.7,
        #                             debug=False)
        trigger_reactions = IA.trigger_reactions

        # no reactions
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "ooooo")
        rep = await trigger_reactions(message)
        self.assertIs(rep, False)
        ctx.assert_sent()

        # no matching reactions
        reac = bdd.Reaction(reponse="non")
        bdd.Trigger(trigger="oui", reaction=reac).add(reac)
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "ooo")
        rep = await trigger_reactions(message)
        self.assertIs(rep, False)
        ctx.assert_sent()

        # lower sensi -> matching
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "ooo")
        rep = await trigger_reactions(message, sensi=0.3)
        self.assertIs(rep, True)
        ctx.assert_sent("non")

        # command
        reac.reponse = f"{IA.MARK_CMD}commandz"
        reac.update()
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "oui")
        with mock.patch("lgrez.config.bot.process_commands") as pc_patch:
            rep = await trigger_reactions(message)
        self.assertIs(rep, True)
        self.assertEqual(message.content, "!commandz")
        ctx.assert_sent()
        pc_patch.assert_called_once_with(message)

        # command - other prefix
        reac.reponse = f"{IA.MARK_CMD}commandz"
        reac.update()
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "oui")
        config.bot.command_prefix = "PREFix"
        with mock.patch("lgrez.config.bot.process_commands") as pc_patch:
            rep = await trigger_reactions(message)
        config.bot.command_prefix = "!"
        self.assertIs(rep, True)
        self.assertEqual(message.content, "PREFixcommandz")
        ctx.assert_sent()
        pc_patch.assert_called_once_with(message)

        # react
        reac.reponse = f"{IA.MARK_REACT}bon."
        reac.update()
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "oui")
        rep = await trigger_reactions(message)
        self.assertIs(rep, True)
        ctx.assert_sent()
        message.add_reaction.assert_called_once_with("bon.")

        # several commands / reactions / texts
        reac.reponse = (f"mezzage{IA.MARK_THEN}{IA.MARK_REACT}bon."
                        f"{IA.MARK_THEN}{IA.MARK_CMD}ouizz"
                        f"{IA.MARK_THEN}aha!{IA.MARK_THEN}{IA.MARK_REACT}krr")
        reac.update()
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "oui")
        with mock.patch("lgrez.config.bot.process_commands") as pc_patch:
            rep = await trigger_reactions(message)
        self.assertIs(rep, True)
        ctx.assert_sent("mezzage", "aha!")
        self.assertEqual(message.add_reaction.mock_calls,
                         [mock.call("bon."), mock.call("krr")])
        self.assertEqual(message.content, "!ouizz")
        pc_patch.assert_called_once_with(message)

        # OR
        reac.reponse = (f"a{IA.MARK_OR}b")
        reac.update()
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "oui")
        na = 0
        nb = 0
        for i in range(1000):
            await trigger_reactions(message)
            sent = ctx.send.call_args.args[0]
            if sent == "a":
                na += 1
            elif sent == "b":
                nb += 1
            else:
                raise AssertionError("Sent something else!")
        self.assertGreater(na, 400)
        self.assertGreater(nb, 400)


    @mock.patch("lgrez.features.IA.trigger_reactions")
    async def test_trigger_sub_reactions(self, tr_patch):
        """Unit tests for IA.trigger_sub_reactions function."""
        # async def trigger_sub_reactions(message, sensi=0.9, debug=False)
        trigger_sub_reactions = IA.trigger_sub_reactions

        # empty message
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "")
        rep = await trigger_sub_reactions(message)
        self.assertIs(rep, False)
        tr_patch.assert_not_called()
        ctx.assert_sent()

        # one word
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "boo")
        rep = await trigger_sub_reactions(message)
        self.assertIs(rep, False)
        tr_patch.assert_not_called()
        ctx.assert_sent()

        # <= 4-length words
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "boo a baaz tok paaf")
        rep = await trigger_sub_reactions(message)
        self.assertIs(rep, False)
        tr_patch.assert_not_called()
        ctx.assert_sent()

        # one > 4-length word, return False
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "boo a baazA tok paaf")
        tr_patch.return_value = False
        rep = await trigger_sub_reactions(message)
        self.assertIs(rep, False)
        tr_patch.assert_called_once_with(message, chain="baazA",
                                         sensi=0.9, debug=False)
        ctx.assert_sent()
        tr_patch.reset_mock()

        # one > 4-length word, return True (and debug True)
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "boo a baazA tok paaf")
        tr_patch.return_value = True
        rep = await trigger_sub_reactions(message, debug=True)
        self.assertIs(rep, True)
        tr_patch.assert_called_once_with(message, chain="baazA",
                                         sensi=0.9, debug=True)
        ctx.assert_sent()
        tr_patch.reset_mock()

        # several > 4-length words, all False
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "book@! a baazA tok pa "
                                            "frooookpoozpzpzpzppa")
        tr_patch.side_effect = [False, False, False]
        rep = await trigger_sub_reactions(message)
        self.assertIs(rep, False)
        self.assertEqual(tr_patch.call_args_list, [
            mock.call(message, chain="frooookpoozpzpzpzppa", sensi=0.9,
                      debug=False),
            mock.call(message, chain="book@!", sensi=0.9, debug=False),
            mock.call(message, chain="baazA", sensi=0.9, debug=False),
        ])
        ctx.assert_sent()
        tr_patch.reset_mock()

        # several > 4-length words, second True
        ctx = mock_discord.get_ctx(None)
        message = mock_discord.message(ctx, "book@! a baazA tok pa "
                                            "frooookpoozpzpzpzppa")
        tr_patch.side_effect = [False, True]
        rep = await trigger_sub_reactions(message)
        self.assertIs(rep, True)
        self.assertEqual(tr_patch.call_args_list, [
            mock.call(message, chain="frooookpoozpzpzpzppa", sensi=0.9,
                      debug=False),
            mock.call(message, chain="book@!", sensi=0.9, debug=False),
        ])
        ctx.assert_sent()
        tr_patch.reset_mock()


    async def test_trigger_di(self):
        """Unit tests for IA.trigger_di function."""
        # async def trigger_di(message)
        trigger_di = IA.trigger_di
        diprefs = ["di", "dy", "dis ", "dit ", "dis-", "dit-"]
        criprefs = ["cri", "cry", "kri", "kry"]

        samples = {
            "": False,
            "boozabaka": False,
            **{f"{pref}kaboom": "kaboom" for pref in diprefs},
            **{f"{pref}kaé@@m": "KAÉ@@M" for pref in criprefs},
            **{f"aa{pref}ka{pref}oh": f"ka{pref}oh" for pref in diprefs},
        }

        # all samples
        for sample, result in samples.items():
            ctx = mock_discord.get_ctx(None)
            message = mock_discord.message(ctx, sample)
            rep = await trigger_di(message)
            if result is False:
                self.assertIs(rep, False)
                ctx.assert_sent()
            else:
                self.assertIs(rep, True)
                ctx.assert_sent(result)


    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.IA.fetch_tenor")
    async def test_trigger_gif(self, ft_patch):
        """Unit tests for IA.trigger_gif function."""
        # async def trigger_gif(message)
        trigger_gif = IA.trigger_gif
        config.bot.in_fals = [12, 13]

        # not in FALS
        ctx = mock_discord.get_ctx(None)
        ctx.channel.id = 2
        message = mock_discord.message(ctx, "booza")
        rep = await trigger_gif(message)
        self.assertIs(rep, False)
        ft_patch.assert_not_called()
        ctx.assert_sent()

        # in FALS, return None
        ctx = mock_discord.get_ctx(None)
        ctx.channel.id = 12
        message = mock_discord.message(ctx, "booza")
        ft_patch.return_value = None
        rep = await trigger_gif(message)
        self.assertIs(rep, False)
        ft_patch.assert_called_once_with("booza")
        ctx.assert_sent()
        ft_patch.reset_mock()

        # in FALS, return a gif
        ctx = mock_discord.get_ctx(None)
        ctx.channel.id = 12
        message = mock_discord.message(ctx, "booza")
        ft_patch.return_value = "gif://cgénial"
        rep = await trigger_gif(message)
        self.assertIs(rep, True)
        ft_patch.assert_called_once_with("booza")
        ctx.assert_sent("gif://cgénial")
        ft_patch.reset_mock()

        config.bot.in_fals = []


    async def test_trigger_mot_unique(self):
        """Unit tests for IA.trigger_mot_unique function."""
        # async def trigger_mot_unique(message)
        trigger_mot_unique = IA.trigger_mot_unique

        samples = {
            "": False,
            "bla bla": False,
            "boozabaka": "Boozabaka ?",
            "booza-baka": "Booza-baka ?",
            "@!!!": "@!!! ?",
            "http://oui": False,
        }

        # all samples
        for sample, result in samples.items():
            ctx = mock_discord.get_ctx(None)
            message = mock_discord.message(ctx, sample)
            rep = await trigger_mot_unique(message)
            if result is False:
                self.assertIs(rep, False)
                ctx.assert_sent()
            else:
                self.assertIs(rep, True)
                ctx.assert_sent(result)


    async def test_trigger_a_ou_b(self):
        """Unit tests for IA.trigger_a_ou_b function."""
        # async def trigger_a_ou_b(message)
        trigger_a_ou_b = IA.trigger_a_ou_b

        samples = {
            "": False,
            "bla bla": False,
            "boozabaka": False,
            "booza ou baka": "Baka.",
            "a v à @ ou b. !??,!;;!": "B.",
            "a v à @ ou b. !??,!;;!a": "B. !??,!;;!a.",
        }

        # all samples
        for sample, result in samples.items():
            ctx = mock_discord.get_ctx(None)
            message = mock_discord.message(ctx, sample)
            rep = await trigger_a_ou_b(message)
            if result is False:
                self.assertIs(rep, False)
                ctx.assert_sent()
            else:
                self.assertIs(rep, True)
                ctx.assert_sent(result)


    async def test_default(self):
        """Unit tests for IA.default function."""
        # async def default(message)
        default = IA.default

        samples = {
            "": False,
            "bla bla": False,
            "boozabaka": False,
            "booza ou baka": "Baka.",
            "a v à @ ou b. !??,!;;!": "B.",
            "a v à @ ou b. !??,!;;!a": "B. !??,!;;!a.",
        }

        # all samples
        for sample, result in samples.items():
            ctx = mock_discord.get_ctx(None)
            message = mock_discord.message(ctx, sample)
            rep = await default(message)
            self.assertIs(rep, True)


    @mock.patch("lgrez.features.IA.trigger_at_mj")
    @mock.patch("lgrez.features.IA.trigger_gif")
    @mock.patch("lgrez.features.IA.trigger_roles")
    @mock.patch("lgrez.features.IA.trigger_reactions")
    @mock.patch("lgrez.features.IA.trigger_sub_reactions")
    @mock.patch("lgrez.features.IA.trigger_a_ou_b")
    @mock.patch("lgrez.features.IA.trigger_di")
    @mock.patch("lgrez.features.IA.trigger_mot_unique")
    @mock.patch("lgrez.features.IA.default")
    async def test_process_IA(self, *patches):
        """Unit tests for IA.process_IA function."""
        # async def process_IA(message, debug=False)
        process_IA = IA.process_IA
        # patches = reverse order of triggers

        # test if none triggered except default, then mot_unique, then...
        for i in range(len(patches)):
            # made reaction = patch[i]
            for ii, patch in enumerate(patches):
                if ii <= i:     # i and lower priority: return True
                    patch.return_value = True
                else:           # higher priority: return False
                    patch.return_value = False

            ctx = mock_discord.get_ctx(None)
            message = mock_discord.message(ctx, "")
            rep = await process_IA(message)
            for ii, patch in enumerate(patches):
                if ii >= i:     # i and higher priority: called
                    patch.assert_called_once()
                    self.assertEqual(patch.call_args.args[0], message)
                else:           # lower priority: not called
                    patch.assert_not_called()
                patch.reset_mock()



class TestGestionIA(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.IA commands."""

    def setUp(self):
        mock_discord.mock_config()
        self.cog = IA.GestionIA(config.bot)

    def tearDown(self):
        mock_discord.unmock_config()


    async def test_stfu(self):
        """Unit tests for !stfu command."""
        # async def stfu(self, ctx, force=None)
        stfu = self.cog.stfu
        config.bot.in_stfu = [12, 13]

        # Not in STFU, no force
        ctx = mock_discord.get_ctx(stfu)
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("je me tais")
        self.assertIn(10, config.bot.in_stfu)

        # In STFU, no force
        ctx = mock_discord.get_ctx(stfu)
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("reparler")
        self.assertNotIn(10, config.bot.in_stfu)

        # Not in STFU, force "on"
        ctx = mock_discord.get_ctx(stfu, force="on")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("je me tais")
        self.assertIn(10, config.bot.in_stfu)

        # In STFU, force "on"
        ctx = mock_discord.get_ctx(stfu, force="on")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertIn(10, config.bot.in_stfu)
        config.bot.in_stfu.remove(10)

        # Not in STFU, force "start"
        ctx = mock_discord.get_ctx(stfu, force="start")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("je me tais")
        self.assertIn(10, config.bot.in_stfu)

        # In STFU, force "start"
        ctx = mock_discord.get_ctx(stfu, force="start")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertIn(10, config.bot.in_stfu)
        config.bot.in_stfu.remove(10)

        # Not in STFU, force "off"
        ctx = mock_discord.get_ctx(stfu, force="off")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertNotIn(10, config.bot.in_stfu)

        # In STFU, force "off"
        config.bot.in_stfu.append(10)
        ctx = mock_discord.get_ctx(stfu, force="off")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("reparler")
        self.assertNotIn(10, config.bot.in_stfu)

        # Not in STFU, force "stop"
        ctx = mock_discord.get_ctx(stfu, force="stop")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertNotIn(10, config.bot.in_stfu)

        # In STFU, force "stop"
        config.bot.in_stfu.append(10)
        ctx = mock_discord.get_ctx(stfu, force="stop")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("reparler")
        self.assertNotIn(10, config.bot.in_stfu)

        # Not in STFU, force autre
        ctx = mock_discord.get_ctx(stfu, force="bzzt")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertIn(10, config.bot.in_stfu)

        # In STFU, force autre
        ctx = mock_discord.get_ctx(stfu, force="bzzt")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertNotIn(10, config.bot.in_stfu)


    async def test_fals(self):
        """Unit tests for !fals command."""
        # async def fals(self, ctx, force=None)
        fals = self.cog.fals
        config.bot.in_fals = [12, 13]

        # Not in FALS, no force
        ctx = mock_discord.get_ctx(fals)
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("saucisse")
        self.assertIn(10, config.bot.in_fals)

        # In FALS, no force
        ctx = mock_discord.get_ctx(fals)
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("pas abuser")
        self.assertNotIn(10, config.bot.in_fals)

        # Not in FALS, force "on"
        ctx = mock_discord.get_ctx(fals, force="on")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("saucisse")
        self.assertIn(10, config.bot.in_fals)

        # In FALS, force "on"
        ctx = mock_discord.get_ctx(fals, force="on")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertIn(10, config.bot.in_fals)
        config.bot.in_fals.remove(10)

        # Not in FALS, force "start"
        ctx = mock_discord.get_ctx(fals, force="start")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("saucisse")
        self.assertIn(10, config.bot.in_fals)

        # In FALS, force "start"
        ctx = mock_discord.get_ctx(fals, force="start")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertIn(10, config.bot.in_fals)
        config.bot.in_fals.remove(10)

        # Not in FALS, force "off"
        ctx = mock_discord.get_ctx(fals, force="off")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertNotIn(10, config.bot.in_fals)

        # In FALS, force "off"
        config.bot.in_fals.append(10)
        ctx = mock_discord.get_ctx(fals, force="off")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("pas abuser")
        self.assertNotIn(10, config.bot.in_fals)

        # Not in FALS, force "stop"
        ctx = mock_discord.get_ctx(fals, force="stop")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertNotIn(10, config.bot.in_fals)

        # In FALS, force "stop"
        config.bot.in_fals.append(10)
        ctx = mock_discord.get_ctx(fals, force="stop")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent("pas abuser")
        self.assertNotIn(10, config.bot.in_fals)

        # Not in FALS, force autre
        ctx = mock_discord.get_ctx(fals, force="bzzt")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertIn(10, config.bot.in_fals)

        # In FALS, force autre
        ctx = mock_discord.get_ctx(fals, force="bzzt")
        ctx.channel.id = 10
        await ctx.invoke()
        ctx.assert_sent()
        self.assertNotIn(10, config.bot.in_fals)


    @mock.patch("lgrez.features.IA.process_IA")
    async def test_react(self, pia_patch):
        """Unit tests for !react command."""
        # async def react(self, ctx, *, trigger)
        react = self.cog.react

        class SaveContent:
            content = None
            @classmethod
            def save(cls, msg, debug=None):
                cls.content = msg.content

        pia_patch.side_effect = SaveContent.save

        # non-MJ
        ctx = mock_discord.get_ctx(react, trigger="booz")
        ctx.author.top_role = config.Role.joueur_en_vie
        ctx.message.content = "original"
        await ctx.invoke()
        pia_patch.assert_called_once_with(ctx.message, debug=False)
        self.assertEqual(SaveContent.content, "booz")
        self.assertEqual(ctx.message.content, "original")
        ctx.assert_sent()
        pia_patch.reset_mock()

        # MJ
        ctx = mock_discord.get_ctx(react, trigger="booz")
        ctx.author.top_role = config.Role.mj
        ctx.message.content = "original"
        await ctx.invoke()
        pia_patch.assert_called_once_with(ctx.message, debug=True)
        self.assertEqual(SaveContent.content, "booz")
        self.assertEqual(ctx.message.content, "original")
        ctx.assert_sent()


    @mock.patch("lgrez.features.IA.fetch_tenor")
    async def test_reactfals(self, ft_patch):
        """Unit tests for !reactfals command."""
        # async def reactfals(self, ctx, *, trigger)
        reactfals = self.cog.reactfals

        # no GIF
        ctx = mock_discord.get_ctx(reactfals, trigger="booz")
        ft_patch.return_value = None
        await ctx.invoke()
        ft_patch.assert_called_once_with("booz")
        ctx.assert_sent("Palaref")
        ft_patch.reset_mock()

        # MJ
        ctx = mock_discord.get_ctx(reactfals, trigger="booz")
        ft_patch.return_value = "giiiif"
        await ctx.invoke()
        ft_patch.assert_called_once_with("booz")
        ctx.assert_sent("giiiif")
        ft_patch.reset_mock()


    @mock_bdd.patch_db
    @mock.patch("lgrez.features.IA._build_sequence")
    async def test_addIA(self, bs_patch):
        """Unit tests for !addIA command."""
        # async def addIA(self, ctx, *, triggers=None)
        addIA = self.cog.addIA

        # no triggers, simple gave
        ctx = mock_discord.get_ctx(addIA)
        bs_patch.return_value = "s€quenZ"
        with mock_discord.interact(
            ("wait_for_message_here", mock_discord.message(ctx, "trigzz"))
        ):
            await ctx.invoke()
        bs_patch.assert_called_once_with(ctx)
        ctx.assert_sent("déclencheurs", "trigzz", "s€quenZ", "ajoutée en base")
        trig = bdd.Trigger.query.one()
        reac = bdd.Reaction.query.one()
        self.assertEqual(trig.trigger, "trigzz")
        self.assertEqual(trig.reaction, reac)
        self.assertEqual(reac.reponse, "s€quenZ")
        bdd.Trigger.delete(trig, reac)
        bs_patch.reset_mock()

        # complex triggers
        triggers = "trigger\n\nTR1; trégèrçù; trÉgÈrÇÙ   ;;  ❤\nah!"
        formated_trgz = ["trigger", "tr1", "tregercu", "tregercu", "❤", "ah!"]
        ctx = mock_discord.get_ctx(addIA, triggers=triggers)
        bs_patch.return_value = "s€quenZ"
        await ctx.invoke()
        bs_patch.assert_called_once_with(ctx)
        ctx.assert_sent(formated_trgz, "s€quenZ", "ajoutée en base")
        trigs = bdd.Trigger.query.all()
        reac = bdd.Reaction.query.one()
        self.assertEqual(len(trigs), 5)
        self.assertEqual({trig.trigger for trig in trigs}, set(formated_trgz))
        self.assertEqual([trig.reaction for trig in trigs], [reac]*5)
        self.assertEqual(reac.reponse, "s€quenZ")
        bs_patch.reset_mock()


    @mock_bdd.patch_db
    async def test_listIA(self):
        """Unit tests for !listIA command."""
        # async def listIA(self, ctx, trigger=None, sensi=0.5)
        listIA = self.cog.listIA

        reacs = [
            bdd.Reaction(reponse="s1quenZ"),
            bdd.Reaction(reponse="s€quenZ2"),
            bdd.Reaction(reponse="s€qu3enZ"),
            bdd.Reaction(reponse="s€4quenZ"),
            bdd.Reaction(reponse="s€que5nZ"*1000),
        ]
        triggers = [
            bdd.Trigger(trigger="trigz1", reaction=reacs[0]),
            bdd.Trigger(trigger="tr1gZ2", reaction=reacs[0]),
            bdd.Trigger(trigger="tr!Gz3", reaction=reacs[0]),
            bdd.Trigger(trigger="tr!Gz4", reaction=reacs[1]),
            bdd.Trigger(trigger="tr!Gz5", reaction=reacs[1]),
            bdd.Trigger(trigger="tr!Gz6", reaction=reacs[2]),
            bdd.Trigger(trigger="tr!Gz7", reaction=reacs[3]),
            bdd.Trigger(trigger="tr!G88", reaction=reacs[4]),
        ]
        bdd.Reaction.add(*reacs, *triggers)

        # list all
        ctx = mock_discord.get_ctx(listIA)
        await ctx.invoke()
        ctx.assert_sent([reac.reponse[:10] for reac in reacs]
                        + [trig.trigger for trig in triggers] + ["[...]"])
        ctx.assert_not_sent("s€que5nZ"*1000)

        # filter
        ctx = mock_discord.get_ctx(listIA, "1gZ2")
        await ctx.invoke()
        ctx.assert_sent(["tr1gZ2", "s1quenZ"])
        ctx.assert_not_sent(["s€quenZ2", "s€qu3enZ", "s€4quenZ", "s€que5nZ"])

        # filter & sensi
        ctx = mock_discord.get_ctx(listIA, "1gZ2", sensi=0.3)
        await ctx.invoke()
        ctx.assert_sent(["s1quenZ", "s€quenZ2", "s€qu3enZ", "s€4quenZ"])
        ctx.assert_not_sent("s€que5nZ")


    @mock_bdd.patch_db
    @mock.patch("lgrez.features.IA._build_sequence")
    async def test_modifIA(self, bs_patch):
        """Unit tests for !modifIA command."""
        # async def modifIA(self, ctx, *, trigger=None)
        modifIA = self.cog.modifIA

        reac = bdd.Reaction(reponse="s€quenZ")
        triggers = [
            bdd.Trigger(trigger="trigz1", reaction=reac),
            bdd.Trigger(trigger="tr1gZ2", reaction=reac),
            bdd.Trigger(trigger="tr!Gz3", reaction=reac),
        ]
        bdd.Reaction.add(reac, *triggers)

        # nothing found
        ctx = mock_discord.get_ctx(modifIA)
        bs_patch.return_value = "s€quenZ"
        with mock_discord.interact(
            ("wait_for_message_here", mock_discord.message(ctx, "ooooo"))
        ):
            await ctx.invoke()
        ctx.assert_sent("déclencheur", "Rien trouvé")
        bs_patch.assert_not_called()

        # modif triggers
        ctx = mock_discord.get_ctx(modifIA, trigger="trigz")
        with mock_discord.interact(
            ("wait_for_react_clic", 1),         # modif triggers
            ("wait_for_react_clic", "ahbon"),   # trigger 4
            ("wait_for_react_clic", "3"),       # delete trigger 3
            ("wait_for_react_clic", "0"),       # end
        ):
            await ctx.invoke()
        ctx.assert_sent(
            ["trigz1", "tr1gZ2", "tr!Gz3", "s€quenZ"], "Modifier",
            ["Supprimer", "ajouter", "trigz1", "tr1gZ2", "tr!Gz3"],
            ["Supprimer", "ajouter", "trigz1", "tr1gZ2", "tr!Gz3", "ahbon"],
            ["Supprimer", "ajouter", "trigz1", "tr1gZ2", "ahbon"],
            "Fini",
        )
        self.assertEqual(bdd.Reaction.query.one(), reac)
        self.assertEqual({trig.trigger for trig in reac.triggers},
                         {"trigz1", "tr1gZ2", "ahbon"})
        self.assertEqual(reac.reponse, "s€quenZ")
        bs_patch.assert_not_called()

        # delete all triggers
        ctx = mock_discord.get_ctx(modifIA, trigger="trigz")
        with mock_discord.interact(
            ("wait_for_react_clic", 1),         # modif triggers
            ("wait_for_react_clic", "3"),       # delete trigger 3
            ("wait_for_react_clic", "1"),       # delete trigger 1
            ("wait_for_react_clic", "1"),       # delete trigger 1
            ("wait_for_react_clic", "0"),       # end
        ):
            await ctx.invoke()
        ctx.assert_sent(
            ["trigz1", "tr1gZ2", "ahbon", "s€quenZ"], "Modifier",
            ["Supprimer", "ajouter", "trigz1", "tr1gZ2", "ahbon"],
            ["Supprimer", "ajouter", "trigz1", "tr1gZ2"],
            ["Supprimer", "ajouter", "tr1gZ2"],
            ["Supprimer", "ajouter"],
            "suppression",
        )
        self.assertEqual(len(bdd.Reaction.query.all()), 0)
        self.assertEqual(len(bdd.Trigger.query.all()), 0)
        bs_patch.assert_not_called()

        # modif sequence (simple)
        reac = bdd.Reaction(reponse="s€quenZ")
        triggers = [
            bdd.Trigger(trigger="trigz1", reaction=reac),
            bdd.Trigger(trigger="tr1gZ2", reaction=reac),
            bdd.Trigger(trigger="tr!Gz3", reaction=reac),
        ]
        bdd.Reaction.add(reac, *triggers)

        ctx = mock_discord.get_ctx(modifIA, trigger="trigz")
        bs_patch.return_value = "new_s€qu€nZ"
        with mock_discord.interact(
            ("wait_for_react_clic", 2),         # modif sequence
        ):
            await ctx.invoke()
        ctx.assert_sent(
            ["trigz1", "tr1gZ2", "tr!Gz3", "s€quenZ"], "Modifier",
            "Fini",
        )
        self.assertEqual(bdd.Reaction.query.one(), reac)
        self.assertEqual({trig.trigger for trig in reac.triggers},
                         {"trigz1", "tr1gZ2", "tr!Gz3"})
        self.assertEqual(reac.reponse, "new_s€qu€nZ")
        bs_patch.assert_called_once_with(ctx)
        bs_patch.reset_mock()

        # modif sequence (complex)
        reac.reponse = f"s€quenZ{IA.MARK_THEN}oui"
        reac.update()

        ctx = mock_discord.get_ctx(modifIA, trigger="trigz")
        bs_patch.return_value = "new_s€qu€nZ"
        with mock_discord.interact(
            ("wait_for_react_clic", 2),         # modif sequence
        ):
            await ctx.invoke()
        ctx.assert_sent(
            ["trigz1", "tr1gZ2", "tr!Gz3", "s€quenZ"], "Modifier",
            "modifiée rapidement",
            "Fini",
        )
        self.assertEqual(bdd.Reaction.query.one(), reac)
        self.assertEqual({trig.trigger for trig in reac.triggers},
                         {"trigz1", "tr1gZ2", "tr!Gz3"})
        self.assertEqual(reac.reponse, "new_s€qu€nZ")
        bs_patch.assert_called_once_with(ctx)
        bs_patch.reset_mock()

        # delete sequence
        ctx = mock_discord.get_ctx(modifIA, trigger="trigz")
        with mock_discord.interact(
            ("wait_for_react_clic", 0),         # dekete sequence
        ):
            await ctx.invoke()
        ctx.assert_sent(
            ["trigz1", "tr1gZ2", "tr!Gz3", "new_s€qu€nZ"], "Modifier",
            "Fini",
        )
        self.assertEqual(len(bdd.Reaction.query.all()), 0)
        self.assertEqual(len(bdd.Trigger.query.all()), 0)
        bs_patch.assert_not_called()



import asyncio
import unittest
from unittest import mock

import discord

from lgrez import config
from lgrez.blocs import realshell
from test import mock_discord


class TestRealShell(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.realshell.RealShell class."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    @mock.patch("asyncode.AsyncInteractiveConsole.__init__")
    def test___init__(self, aic_patch):
        """Unit tests for RealShell.__init__ method."""
        # def __init__(self, channel, locals={}, filename="!shell")
        __init__ = realshell.RealShell.__init__
        # simple
        slf = mock.Mock(realshell.RealShell)
        channel = mock.Mock()
        __init__(slf, channel)
        self.assertEqual(slf.bot, config.bot)
        self.assertEqual(slf.channel, channel)
        aic_patch.assert_called_once_with({"_shell": slf}, "!shell")
        aic_patch.reset_mock()
        # complex
        slf = mock.Mock(realshell.RealShell)
        channel, filename = mock.Mock(), mock.Mock()
        __init__(slf, channel, locals={"a": 1, "b": "oh"}, filename=filename)
        self.assertEqual(slf.bot, config.bot)
        self.assertEqual(slf.channel, channel)
        aic_patch.assert_called_once_with({"a": 1, "b": "oh", "_shell": slf},
                                          filename)


    @mock.patch("asyncode.AsyncInteractiveConsole.interact")
    @mock.patch("lgrez.blocs.realshell._py_bloc")
    async def test_interact(self, bloc_patch, aic_patch):
        """Unit tests for RealShell.interact method."""
        # async def interact(self, banner=None, exitmsg="Byebye!")
        interact = realshell.RealShell.interact
        bloc_patch.side_effect = lambda data: f"<!<{data}>!>"

        # simple
        channel = mock_discord.chan("tezt")
        slf = mock.Mock(realshell.RealShell, channel=channel)
        await interact(slf)
        mock_discord.assert_sent(channel, "Launching Python shell")
        aic_patch.assert_called_once_with(mock.ANY, "Byebye!")
        banner = aic_patch.call_args.args[0]
        self.assertIn("Discord implementation", banner)
        slf.message.clear_reactions.assert_called_once()
        aic_patch.reset_mock()

        # complex
        channel = mock_discord.chan("tezt")
        slf = mock.Mock(realshell.RealShell, channel=channel)
        await interact(slf, banner="LULZ", exitmsg="allo?")
        mock_discord.assert_sent(channel, "Launching Python shell")
        aic_patch.assert_called_once_with("LULZ", "allo?")
        slf.message.clear_reactions.assert_called_once()
        aic_patch.reset_mock()

        # simple, exited
        aic_patch.side_effect = SystemExit
        channel = mock_discord.chan("tezt")
        slf = mock.Mock(realshell.RealShell, channel=channel)
        with self.assertRaises(realshell.RealShellExit):
            await interact(slf)
        mock_discord.assert_sent(channel, "Launching Python shell")
        aic_patch.assert_called_once_with(mock.ANY, "Byebye!")
        slf.write.assert_called_once_with(f"\nByebye!", refresh=True,
                                          show_caret=False)
        slf.message.clear_reactions.assert_called_once()
        aic_patch.reset_mock()

        # complex, exited
        channel = mock_discord.chan("tezt")
        slf = mock.Mock(realshell.RealShell, channel=channel)
        aic_patch.side_effect = SystemExit
        with self.assertRaises(realshell.RealShellExit):
            await interact(slf, banner="LULZ", exitmsg="allo?")
        mock_discord.assert_sent(channel, "Launching Python shell")
        aic_patch.assert_called_once_with("LULZ", "allo?")
        slf.write.assert_called_once_with(f"\nallo?", refresh=True,
                                          show_caret=False)
        slf.message.clear_reactions.assert_called_once()


    async def test_control(self):
        """Unit tests for RealShell.control method."""
        # async def control(self, self, command, times=1)
        control = realshell.RealShell.control

        # bad command
        slf = mock.Mock(realshell.RealShell)
        with self.assertRaises(ValueError):
            await control(slf, "uiz")

        samples = {
            # (COMMAND, indent, times): (indent, result, write call)
            (realshell.RSCommand.HOME,  0, None): (0,  None, None),
            (realshell.RSCommand.HOME,  3, None): (0,  None, ""),
            (realshell.RSCommand.UNTAB, 0, None): (0,  None, None),
            (realshell.RSCommand.UNTAB, 0, 3):    (0,  None, None),
            (realshell.RSCommand.UNTAB, 5, None): (4,  None, ""),
            (realshell.RSCommand.UNTAB, 5, 3):    (2,  None, ""),
            (realshell.RSCommand.UNTAB, 5, 7):    (0,  None, ""),
            (realshell.RSCommand.TAB,   0, None): (1,  None, ""),
            (realshell.RSCommand.TAB,   0, 3):    (3,  None, ""),
            (realshell.RSCommand.TAB,   5, None): (6,  None, ""),
            (realshell.RSCommand.TAB,   5, 3):    (8,  None, ""),
            (realshell.RSCommand.TAB,   5, 7):    (12, None, ""),
            (realshell.RSCommand.ENTER, 0, None): (0,  "", None),
            (realshell.RSCommand.ENTER, 0, 3):    (0,  "", None),
            (realshell.RSCommand.ENTER, 5, None): (5,  "", None),
            (realshell.RSCommand.ENTER, 5, 3):    (5,  "", None),
            (realshell.RSCommand.ENTER, 5, 7):    (5,  "", None),
        }

        for (COMM, ind, times), (ind_res, res, wrc) in samples.items():
            slf = mock.Mock(realshell.RealShell, indent=ind)
            if times is None:
                ret = await control(slf, COMM)
            else:
                ret = await control(slf, COMM.value, times=times)
                # on teste avec .value de temps en temps aussi
            self.assertEqual(slf.indent, ind_res)
            self.assertEqual(ret, res)
            if wrc is None:
                slf.write.assert_not_called()
            else:
                slf.write.assert_called_once_with(wrc, refresh=True)

        samples = {
            # (COMMAND, indent, times):(indent
            (realshell.RSCommand.EOF,   0, None): 0,
            (realshell.RSCommand.EOF,   0, 3):    0,
            (realshell.RSCommand.EOF,   5, None): 5,
            (realshell.RSCommand.EOF,   5, 3):    5,
            (realshell.RSCommand.EOF,   5, 3):    5,
        }

        for (COMM, ind, times), ind_res in samples.items():
            slf = mock.Mock(realshell.RealShell, indent=ind)
            with self.assertRaises(realshell.RealShellExit):
                if times is None:
                     await control(slf, COMM)
                else:
                     await control(slf, COMM.value, times=times)
                # on teste avec .value de temps en temps aussi
            self.assertEqual(slf.indent, ind_res)
            slf.write.assert_not_called()


    async def test__check_reacts(self):
        """Unit tests for RealShell._check_reacts method."""
        # async def _check_reacts(self)
        _check_reacts = realshell.RealShell._check_reacts

        class AsyncIter():
            def __init__(self, _iter):
                self._iter = iter(_iter)

            async def __anext__(self):
                return next(self._iter)
        class ReactUsersRV(mock.Mock):
            def __iter__(self):
                return iter(self.users)
            def __aiter__(self):
                return AsyncIter(self.users)
        class ReactUsersMocker(mock.MagicMock):
            @property
            def return_value(self):
                return ReactUsersRV(users=self.users)

        # no reacts
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(),
                        message=mock.Mock(discord.Message))
        slf.message.reactions = []
        ret = await _check_reacts(slf)
        slf.message.add_reaction.assert_has_calls(
            [mock.call(command.value) for command in realshell.RSCommand])
        self.assertIs(False, ret)

        # reacts before, none clicked
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(),
                        message=mock.Mock(discord.Message))
        reacts = [mock.Mock(discord.Reaction,
                            **{"emoji.name": command.value}, count=1)
                  for command in realshell.RSCommand]
        slf.message.reactions = reacts
        ret = await _check_reacts(slf)
        slf.message.add_reaction.assert_not_called()
        for react in reacts:
            react.remove.assert_not_called()
        slf.control.assert_not_called()
        self.assertIs(False, ret)

        # reacts before, TAB clicked
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(),
                        message=mock.Mock(discord.Message))
        user = mock.Mock()
        reacts = [
            mock.Mock(discord.Reaction, **{"emoji.name": command.value},
                      count=2 if command == realshell.RSCommand.TAB else 1,
                      users=ReactUsersMocker(users=[slf.bot.user] + ([user]
                            if command == realshell.RSCommand.TAB else [])))
            for command in realshell.RSCommand
        ]
        slf.message.reactions = reacts
        ret = await _check_reacts(slf)
        slf.message.add_reaction.assert_not_called()
        for react in reacts:
            if react.emoji.name == realshell.RSCommand.TAB.value:
                react.remove.assert_called_once_with(user)
            else:
                react.remove.assert_not_called()
        slf.control.assert_called_once_with(realshell.RSCommand.TAB)
        slf.control.reset_mock()
        self.assertEqual(ret, (True, slf.control.return_value))

        # reacts before, more clicked -> one treated only
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(),
                        message=mock.Mock(discord.Message))
        reacts = [
            mock.Mock(discord.Reaction, **{"emoji.name": command.value},
                      count=2 if i > 2 else 1,
                      users=ReactUsersMocker(users=[slf.bot.user] +
                            ([user] if i > 2 else [])))
            for i, command in enumerate(realshell.RSCommand)
        ]
        slf.message.reactions = reacts
        ret = await _check_reacts(slf)
        slf.message.add_reaction.assert_not_called()
        for react in reacts:
            if react == reacts[3]:
                react.remove.assert_called_once()
            else:
                react.remove.assert_not_called()
        slf.control.assert_called_once_with(
            realshell.RSCommand(reacts[3].emoji.name))
        slf.control.reset_mock()
        self.assertEqual(ret, (True, slf.control.return_value))


    @mock.patch("lgrez.blocs.tools.wait_for_message", new_callable=mock.Mock)
    @mock.patch("asyncio.create_task")
    @mock.patch("asyncio.wait")
    async def test_get_order(self, aw_patch, act_patch, wfm_patch):
        """Unit tests for RealShell.get_order method."""
        # async def get_order(self)
        get_order = realshell.RealShell.get_order
        RSCommand = realshell.RSCommand

        # clic anticipé
        slf = mock.Mock(realshell.RealShell)
        ctrl_ret = mock.Mock()
        slf._check_reacts.return_value = (True, ctrl_ret)
        ret = await get_order(slf)
        slf._check_reacts.assert_called_once()
        self.assertEqual(ret, ctrl_ret)
        aw_patch.assert_not_called()
        act_patch.assert_not_called()
        wfm_patch.assert_not_called()

        # clic on TAB react
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(),
                        channel=mock.Mock(),
                        message=mock.Mock(discord.Message))
        slf._check_reacts.return_value = False       # pas de clic anticipé
        slf.message.reactions = []
        tasks = [mock.Mock(), mock.Mock()]      # react, message
        tasks[0].result.return_value.emoji.name = RSCommand.TAB.value
        act_patch.side_effect = tasks
        aw_patch.return_value = {tasks[0]}, {tasks[1]}

        ret = await get_order(slf)
        slf._check_reacts.assert_called_once()
        slf.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                 check=mock.ANY)
        wfm_patch.assert_called_once_with(trigger_on_commands=True,
                                          check=mock.ANY)
        react_check = slf.bot.wait_for.call_args.kwargs["check"]
        message_check = wfm_patch.call_args.kwargs["check"]
        self.assertEqual(act_patch.call_count, 2)
        act_patch.assert_has_calls([mock.call(slf.bot.wait_for.return_value),
                                    mock.call(wfm_patch.return_value)])
        aw_patch.assert_called_once_with(set(tasks),
                                         return_when=asyncio.FIRST_COMPLETED)
        tasks[1].cancel.assert_called_once()
        tasks[0].cancel.assert_not_called()
        p = tasks[0].result.return_value
        slf.message.remove_reaction.assert_called_once_with(p.emoji, p.member)
        slf.control.assert_called_once_with(RSCommand.TAB)
        self.assertEqual(ret, slf.control.return_value)
        aw_patch.reset_mock()
        act_patch.reset_mock()
        wfm_patch.reset_mock()

        # interlude: check checks
        slf.message.id = 15
        slf.bot.user.id = 33
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=31)))
        self.assertFalse(react_check(mock.Mock(message_id=11, user_id=33)))
        self.assertTrue(react_check(mock.Mock(message_id=15, user_id=31)))
        self.assertFalse(react_check(mock.Mock(message_id=15, user_id=33)))
        chan = slf.channel
        auth = slf.bot.user
        self.assertFalse(message_check(mock.Mock(channel="oh", author="meh")))
        self.assertFalse(message_check(mock.Mock(channel="oh", author=auth)))
        self.assertTrue(message_check(mock.Mock(channel=chan, author="meh")))
        self.assertFalse(message_check(mock.Mock(channel=chan, author=auth)))

        # message response
        samples = {
            # input -> (return value, control calls)
            "bzzt":     ("bzzt",    []),
            "<":        (None,      [mock.call(RSCommand.UNTAB, times=1)]),
            "<<<":      (None,      [mock.call(RSCommand.UNTAB, times=3)]),
            ">":        (None,      [mock.call(RSCommand.TAB, times=1)]),
            ">>>":      (None,      [mock.call(RSCommand.TAB, times=3)]),
            "<<>>":     (">>",      [mock.call(RSCommand.UNTAB, times=2)]),
            ">><<":     ("<<",      [mock.call(RSCommand.TAB, times=2)]),
            "<bzzt":    ("bzzt",    [mock.call(RSCommand.UNTAB, times=1)]),
            "<<<bzzt":  ("bzzt",    [mock.call(RSCommand.UNTAB, times=3)]),
            ">bzzt":    ("bzzt",    [mock.call(RSCommand.TAB, times=1)]),
            ">>>bzzt":  ("bzzt",    [mock.call(RSCommand.TAB, times=3)]),
            "<<>>bzzt": (">>bzzt",  [mock.call(RSCommand.UNTAB, times=2)]),
            "<=":       (...,       [mock.call(RSCommand.UNTAB, times=1),
                                     mock.call(RSCommand.ENTER)]),
            "<<<=":     (...,       [mock.call(RSCommand.UNTAB, times=3),
                                     mock.call(RSCommand.ENTER)]),
            ">=":       (...,       [mock.call(RSCommand.TAB, times=1),
                                     mock.call(RSCommand.ENTER)]),
            ">>>=":     (...,       [mock.call(RSCommand.TAB, times=3),
                                     mock.call(RSCommand.ENTER)]),
            "<<>>=":    (">>=",     [mock.call(RSCommand.UNTAB, times=2)]),
            "<a=b":     ("a=b",     [mock.call(RSCommand.UNTAB, times=1)]),
            ">>a=b":    ("a=b",     [mock.call(RSCommand.TAB, times=2)]),
            "a<=b":     ("a<=b",    []),
            "a>=b>":    ("a>=b>",   []),
            "a>=b><<":  ("a>=b><<", []),
            "<==":      ("==",      [mock.call(RSCommand.UNTAB, times=1)]),
            "==":       ("==",      []),
            "=":        (...,       [mock.call(RSCommand.ENTER)]),
        }

        for inpt, (expected, exp_calls) in samples.items():
            slf = mock.Mock(realshell.RealShell, bot=mock.Mock(),
                            channel=mock.Mock(),
                            message=mock.Mock(discord.Message))
            slf._check_reacts.return_value = False    # pas de clic anticipé
            slf.message.reactions = []
            tasks = [mock.Mock(), mock.Mock()]        # react, message
            tasks[1].result.return_value = mock.Mock(discord.Message,
                                                     content=inpt)
            act_patch.side_effect = tasks
            aw_patch.return_value = {tasks[1]}, {tasks[0]}

            ret = await get_order(slf)
            slf._check_reacts.assert_called_once()
            slf.bot.wait_for.assert_called_once_with('raw_reaction_add',
                                                     check=mock.ANY)
            wfm_patch.assert_called_once_with(trigger_on_commands=True,
                                              check=mock.ANY)
            react_check = slf.bot.wait_for.call_args.kwargs["check"]
            message_check = wfm_patch.call_args.kwargs["check"]
            self.assertEqual(act_patch.call_count, 2)
            act_patch.assert_has_calls(
                [mock.call(slf.bot.wait_for.return_value),
                 mock.call(wfm_patch.return_value)])
            aw_patch.assert_called_once_with(
                set(tasks), return_when=asyncio.FIRST_COMPLETED)
            tasks[0].cancel.assert_called_once()
            tasks[1].cancel.assert_not_called()
            mess = tasks[1].result.return_value
            mess.delete.assert_called_once()
            if expected is ...:
                expected = slf.control.return_value
            self.assertEqual(ret, expected, msg=inpt)
            self.assertEqual(slf.control.call_count, len(exp_calls), msg=inpt)
            slf.control.assert_has_calls(exp_calls)
            aw_patch.reset_mock()
            act_patch.reset_mock()
            wfm_patch.reset_mock()


    @mock.patch("lgrez.blocs.realshell._py_bloc")
    @mock.patch("lgrez.blocs.tools.smooth_split")
    async def test_write(self, tss_patch, aic_patch):
        """Unit tests for RealShell.write method."""
        # async def write(self, data, *, refresh=False, show_caret=True)
        write = realshell.RealShell.write
        aic_patch.side_effect = lambda data: f"<!<{data}>!>"

        # basic - non full -> nothing
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(), indent=2,
                        message=mock.Mock(discord.Message, content="<!<oh>!>"),
                        rsbuffer="kook", content="oh")
        await write(slf, "drzzz")
        self.assertEqual(slf.rsbuffer, "kookdrzzz")
        self.assertEqual(slf.indent, 2)
        self.assertEqual(slf.content, "oh")
        slf.message.edit.assert_not_called()
        slf.message.clear_reactions.assert_not_called()
        aic_patch.assert_not_called()
        tss_patch.assert_not_called()

        # refresh - non full
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(), indent=2,
                        message=mock.Mock(discord.Message, content="<!<oh>!>"),
                        rsbuffer="kook", content="oh")
        await write(slf, "drzzz", refresh=True)
        self.assertEqual(slf.rsbuffer, "")
        self.assertEqual(slf.indent, 2)
        self.assertEqual(slf.content, "ohkookdrzzz")
        slf.message.edit.assert_called_once_with(
            content="<!<ohkookdrzzz        ‸>!>")
        slf.message.clear_reactions.assert_not_called()
        tss_patch.assert_not_called()

        # refresh - no caret - non full
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(), indent=2,
                        message=mock.Mock(discord.Message, content="<!<oh>!>"),
                        rsbuffer="kook", content="oh")
        await write(slf, "drzzz", refresh=True, show_caret=False)
        self.assertEqual(slf.rsbuffer, "")
        self.assertEqual(slf.indent, 2)
        self.assertEqual(slf.content, "ohkookdrzzz")
        slf.message.edit.assert_called_once_with(content="<!<ohkookdrzzz>!>")
        slf.message.clear_reactions.assert_not_called()
        tss_patch.assert_not_called()

        # ends by sys.ps1 -> refresh auto - non full
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(), indent=2,
                        message=mock.Mock(discord.Message, content="<!<oh>!>"),
                        rsbuffer="kook", content="oh")
        await write(slf, "drzzz\n>>> ")
        self.assertEqual(slf.rsbuffer, "")
        self.assertEqual(slf.indent, 2)
        self.assertEqual(slf.content, "ohkookdrzzz\n>>> ")
        slf.message.edit.assert_called_once_with(
            content="<!<ohkookdrzzz\n>>>         ‸>!>")
        slf.message.clear_reactions.assert_not_called()
        tss_patch.assert_not_called()

        # ends by sys.ps2 -> refresh auto - non full
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(), indent=2,
                        message=mock.Mock(discord.Message, content="<!<oh>!>"),
                        rsbuffer="kook", content="oh")
        await write(slf, "drzzz\n... ")
        self.assertEqual(slf.rsbuffer, "")
        self.assertEqual(slf.indent, 2)
        self.assertEqual(slf.content, "ohkookdrzzz\n... ")
        slf.message.edit.assert_called_once_with(
            content="<!<ohkookdrzzz\n...         ‸>!>")
        slf.message.clear_reactions.assert_not_called()
        tss_patch.assert_not_called()

        # ends by :\nsys.ps2 -> refresh auto + autoindent - non full
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(), indent=2,
                        message=mock.Mock(discord.Message, content="<!<oh>!>"),
                        rsbuffer="kook", content="oh")
        await write(slf, "drzzz:\n... ")
        self.assertEqual(slf.rsbuffer, "")
        self.assertEqual(slf.indent, 3)
        self.assertEqual(slf.content, "ohkookdrzzz:\n... ")
        slf.message.edit.assert_called_once_with(
            content="<!<ohkookdrzzz:\n...             ‸>!>")
        slf.message.clear_reactions.assert_not_called()
        tss_patch.assert_not_called()

        # refresh - content(2) + buffer(4) + data(x) + insert(9) = 1991 => full
        # x = (1991 - 9 - 4 - 2) = 1976 + 4*494
        chan = mock_discord.chan(".")
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(), indent=2,
                        message=mock.Mock(discord.Message, content="<!<oh>!>"),
                        rsbuffer="kook", content="oh", channel=chan)
        orig_msg = slf.message
        tss_patch.return_value = ["bl1", "bl2", "bl3"]
        await write(slf, "drzZ"*494, refresh=True)
        orig_msg.clear_reactions.assert_called_once()
        self.assertEqual(slf.rsbuffer, "")
        self.assertEqual(slf.indent, 2)
        self.assertEqual(slf.content, "bl3")
        slf.message.edit.assert_not_called()
        tss_patch.assert_called_once_with("kook" + "drzZ" * 494, N=1990)
        self.assertEqual(chan.send.call_count, 3)
        chan.send.assert_has_calls([mock.call("<!<bl1>!>"),
                                    mock.call("<!<bl2>!>"),
                                    mock.call("<!<bl3        ‸>!>")])
        self.assertEqual(slf.message, chan.send.return_value)


    async def test_raw_input(self):
        """Unit tests for RealShell.raw_input method."""
        # async def raw_input(self, prompt="")
        raw_input = realshell.RealShell.raw_input

        # basic
        chan = mock_discord.chan(".")
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(), indent=2)
        slf.get_order.return_value = "oh"
        ret = await raw_input(slf)
        slf.get_order.assert_called_once()
        slf.write.assert_called_once_with("        oh\n", refresh=True,
                                          show_caret=False)
        self.assertEqual(ret, "        oh")

        # prompt + get_order loop
        chan = mock_discord.chan(".")
        slf = mock.Mock(realshell.RealShell, bot=mock.Mock(), indent=2)
        slf.get_order.side_effect = [None, None, ""]
        ret = await raw_input(slf, prompt="klkl")
        self.assertEqual(slf.get_order.call_count, 3)
        self.assertEqual(slf.write.call_count, 2)
        slf.write.assert_has_calls([
            mock.call("klkl", refresh=True),
            mock.call("        \n", refresh=True, show_caret=False),
        ])
        self.assertEqual(ret, "        ")



class TestRealShellFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.realshell functions."""

    @mock.patch("lgrez.blocs.tools.code_bloc")
    def test__py_bloc(self, cb_patch):
        """Unit tests for realshell._py_bloc function."""
        # def _py_bloc(data)
        _py_bloc = realshell._py_bloc
        # unique case
        data = mock.Mock()
        ret = _py_bloc(data)
        cb_patch.assert_called_once_with(data, "py")
        self.assertEqual(ret, cb_patch.return_value)



"""lg-rez / bdd / Enums

Déclaration des énumérations utilisées

"""

import enum


# Définition des enums

class Statut(enum.Enum):
    """:class:`~enum.Enum` représentant le statut RP d'un Joueur.

    Attributes:
        vivant: Le joueur est en vie !
        mort: Le joueur est mort. RIP.
        MV: Le joueur est Mort-Vivant. Pas de chance.
        immortel: Le joueur est Immortel. Si jamais...
    """
    vivant = enum.auto()
    mort = enum.auto()
    MV = enum.auto()
    immortel = enum.auto()


class ActionTrigger(enum.Enum):
    """:class:`~enum.Enum` : Déclencheur de l'ouverture/fermeture d'une Action.

    Attributes:
        temporel: Ouverture/fermeture à heure fixe chaque jour
        delta: Fermeture un délai donné après l'ouverture
        perma: Action utilisable en permanence
        start: Ouverture au lancement du jeu
        auto: Action automatique, fermeture dès l'ouverture
        mort: Ouverture à la mort
        mot_mjs: Ouverture/fermeture au !plot cond
        open_cond: À l'ouverture du vote condamné
        close_cond: À la fermeture du vote condamné
        open_maire: À l'ouverture du vote du maire
        close_maire: À la fermeture du vote du maire
        open_loups: À l'ouverture du vote des loups
        close_loups: À la fermeture du vote des loups
    """
    temporel = enum.auto()
    delta = enum.auto()
    perma = enum.auto()
    start = enum.auto()
    auto = enum.auto()
    mort = enum.auto()
    mot_mjs = enum.auto()
    open_cond = enum.auto()
    close_cond = enum.auto()
    open_maire = enum.auto()
    close_maire = enum.auto()
    open_loups = enum.auto()
    close_loups = enum.auto()

    @classmethod
    def open(cls, vote):
        if not isinstance(vote, Vote):
            vote = Vote[vote]
        return cls[f"open_{vote.name}"]

    @classmethod
    def close(cls, vote):
        if not isinstance(vote, Vote):
            vote = Vote[vote]
        return cls[f"close_{vote.name}"]


class CandidHaroType(enum.Enum):
    """:class:`~enum.Enum` pour distinguer un haro d'une candidature.

    Attributes:
        candidature: Candidature à la mairie
        haro: Haro pour le bûcher
    """
    candidature = enum.auto()
    haro = enum.auto()


class Vote(enum.Enum):
    """:class:`~enum.Enum` représentant les différents votes possibles.

    Attributes:
        cond: Vote pour le condamné du jour
        maire: Vote pour le nouveau maire
        loups: Vote pour la victime des loups
    """
    cond = enum.auto()
    maire = enum.auto()
    loups = enum.auto()


class UtilEtat(enum.Enum):
    """:class:`~enum.Enum` représentant l'état d'une utilisation d'action.

    Attributes:
        ouverte: Action utilisable, pas de décision prise
        remplie: Décision prise, encore possibilité de changer
        validee: Décision prise et bloquée (créneau pour agir fini)
        ignoree: Créneau pour agir fini sans qu'une décision soit prise
        contree: Utilisation validée, mais contrée par un évènement externe
    """
    ouverte = enum.auto()
    remplie = enum.auto()
    validee = enum.auto()
    ignoree = enum.auto()
    contree = enum.auto()


class CibleType(enum.Enum):
    """:class:`~enum.Enum` représentant le type de cible d'une action.

    Attributes:
        joueur: La cible doit être un joueur, vivant ou mort
        vivant: La cible doit être un joueur vivant
        mort: La cible doit être un joueur mort
        role: La cible doit être un rôle existant
        camp: La cible doit être un camp (public) existant
        booleen: Oui/non : sert à ue action sans paramètres, à une option...
        texte: Permet d'entrer un texte libre
    """
    joueur = enum.auto()
    vivant = enum.auto()
    mort = enum.auto()
    role = enum.auto()
    camp = enum.auto()
    booleen = enum.auto()
    texte = enum.auto()



import unittest
from unittest import mock

from lgrez import config, bdd
from lgrez.features import informations
from test import mock_discord, mock_bdd


class TestInformations(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.informations."""

    def setUp(self):
        mock_discord.mock_config()
        self.cog = informations.Informations(config.bot)

    def tearDown(self):
        mock_discord.unmock_config()

    @mock_bdd.patch_db      # Empty database for this method
    async def test_roles(self):
        """Unit tests for !roles command."""
        # async def roles(self, ctx, *, filtre=None)
        roles = self.cog.roles

        # !roles with no roles
        ctx = mock_discord.get_ctx(roles)
        await ctx.invoke()
        ctx.assert_sent("Rôles trouvés :")

        mock_bdd.add_campsroles(300, 300)       # several panes

        # !roles with a lot of roles
        emoji1 = mock.Mock(name="<:emoji1:>")
        emoji1.configure_mock(name="emoji1")
        ctx = mock_discord.get_ctx(roles)
        with mock.patch("lgrez.config.guild.emojis", [emoji1]):
            await ctx.invoke()
        sent = "\n".join(call.args[0] for call in ctx.send.call_args_list)
        self.assertIn("Role1", sent)
        self.assertIn("emoji1", sent)
        self.assertNotIn("emoji2", sent)

        bdd.Role(slug="chz", nom="ChienZ", description_courte="zoo",
                 description_longue="zpp"*12,
                 camp=bdd.Camp.query.get("camp42")).add()

        # !roles <camp>
        ctx = mock_discord.get_ctx(roles, filtre="camp42")
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Role42", ctx.send.call_args.args[0])
        self.assertIn("ChienZ", ctx.send.call_args.args[0])

        # !roles <role>
        ctx = mock_discord.get_ctx(roles, filtre="chz")
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("ChienZ", ctx.send.call_args.args[0])
        self.assertIn("Camp42", ctx.send.call_args.args[0])
        self.assertIn("zoo", ctx.send.call_args.args[0])
        self.assertIn("zppzpp", ctx.send.call_args.args[0])


    @mock_bdd.patch_db      # Empty database for this method
    async def test_rolede(self):
        """Unit tests for !rolede command."""
        # async def rolede(self, ctx, *, cible=None)
        rolede = self.cog.rolede
        mock_bdd.add_campsroles(10, 10)
        bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                   _role_slug="role7").add()

        # !rolede with existing cible
        ctx = mock_discord.get_ctx(rolede, cible="Joueur1")
        await ctx.invoke()
        ctx.assert_sent("Role7")

        # !rolede with no cible
        ctx = mock_discord.get_ctx(rolede)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("zzz")),
                ("wait_for_message_here", ctx.new_message("zzz")),
                ("wait_for_message_here", ctx.new_message("Joueur1"))):
            await ctx.invoke()
        ctx.assert_sent("", "", "", "Role7")


    @mock_bdd.patch_db      # Empty database for this method
    async def test_quiest(self):
        """Unit tests for !quiest command."""
        # async def quiest(self, ctx, *, nomrole)
        quiest = self.cog.quiest
        mock_bdd.add_campsroles(10, 10)
        bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                   _role_slug="role7").add()
        bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2",
                   _role_slug="role8").add()
        bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3",
                   _role_slug="role8").add()
        bdd.Joueur(discord_id=4, chan_id_=41, nom="Joueur4",
                   _role_slug="role8", statut=bdd.Statut.mort).add()

        # !quiest with non-existing nomrole
        ctx = mock_discord.get_ctx(quiest, nomrole="zzzzz")
        await ctx.invoke()
        ctx.assert_sent("")

        # !quiest with existing nomrole
        ctx = mock_discord.get_ctx(quiest, nomrole="role7")
        await ctx.invoke()
        ctx.assert_sent("Joueur1")

        # check several players & dead
        ctx = mock_discord.get_ctx(quiest, nomrole="role8")
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertIn("Joueur3", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur4", ctx.send.call_args.args[0])


    @mock_bdd.patch_db      # Empty database for this method
    async def test_menu(self):
        """Unit tests for !menu command."""
        # async def menu(self, ctx)
        menu = self.cog.menu
        mock_bdd.add_campsroles(10, 10)
        joueur = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                            _role_slug="role7")
        joueur.add()

        # no vote nor actions
        ctx = mock_discord.get_ctx(menu, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("MENU", ctx.send.call_args.args[0])
        self.assertIn("Aucun vote en cours", ctx.send.call_args.args[0])
        self.assertIn("Aucune action en cours", ctx.send.call_args.args[0])

        # all votes, no actions
        joueur.vote_condamne_ = "zoopla"
        joueur.vote_maire_ = "zooplo"
        joueur.vote_loups_ = "zoopli"
        joueur.update()
        ctx = mock_discord.get_ctx(menu, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        sent = ctx.send.call_args.args[0]
        self.assertIn("Vote pour le bûcher en cours", sent)
        self.assertIn("Vote pour le maire en cours", sent)
        self.assertIn("Vote des loups en cours", sent)
        self.assertIn("zoopla", sent)
        self.assertIn("zooplo", sent)
        self.assertIn("zoopli", sent)
        self.assertIn("Aucune action en cours", sent)

        # all votes, one action but closed
        bdd.BaseAction(slug="ouiZ", trigger_debut=bdd.ActionTrigger.perma,
                       trigger_fin=bdd.ActionTrigger.perma).add()
        action = bdd.Action(joueur=joueur, _base_slug="ouiZ")
        action.add()
        ctx = mock_discord.get_ctx(menu, _caller_id=1)
        await ctx.invoke()
        ctx.assert_sent("Aucune action en cours")

        # all votes, one action
        action.decision_ = "neIn"
        ctx = mock_discord.get_ctx(menu, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        sent = ctx.send.call_args.args[0]
        self.assertIn("Action en cours", sent)
        self.assertIn("ouiZ", sent)
        self.assertIn("neIn", sent)

        # all votes, two actions
        bdd.BaseAction(slug="JaJaJa", trigger_debut=bdd.ActionTrigger.mot_mjs,
                       trigger_fin=bdd.ActionTrigger.perma).add()
        bdd.Action(joueur=joueur, _base_slug="JaJaJa", decision_="o0ps").add()
        ctx = mock_discord.get_ctx(menu, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        sent = ctx.send.call_args.args[0]
        self.assertIn("Action en cours", sent)
        self.assertIn("JaJaJa", sent)
        self.assertIn("o0ps", sent)


    @mock_bdd.patch_db      # Empty database for this method
    async def test_infos(self):
        """Unit tests for !infos command."""
        # async def infos(self, ctx)
        infos = self.cog.infos
        mock_bdd.add_campsroles(10, 10)
        joueur = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                            _role_slug="role7")
        joueur.add()

        # no actions
        ctx = mock_discord.get_ctx(infos, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Role7", ctx.send.call_args.args[0])
        self.assertIn("role7", ctx.send.call_args.args[0])
        self.assertIn("Aucune action disponible", ctx.send.call_args.args[0])

        # one action
        bdd.BaseAction(slug="ouiZ", trigger_debut=bdd.ActionTrigger.perma,
                       trigger_fin=bdd.ActionTrigger.perma).add()
        bdd.Action(joueur=joueur, _base_slug="ouiZ").add()
        ctx = mock_discord.get_ctx(infos, _caller_id=1)
        await ctx.invoke()
        ctx.assert_sent("ouiZ")

        # test every triggers
        names = []
        for trigger in bdd.ActionTrigger:
            name = f"test{trigger.name}"
            bdd.BaseAction(slug=name, trigger_debut=trigger,
                           trigger_fin=trigger).add()
            bdd.Action(joueur=joueur, _base_slug=name).add()
            names.append(name)
        ctx = mock_discord.get_ctx(infos, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        for name in names:
            self.assertIn(name, ctx.send.call_args.args[0])


    @mock_bdd.patch_db      # Empty database for this method
    async def test_vivants_morts(self):
        """Unit tests for !vivants and !morts commands."""
        # async def vivants(self, ctx)
        # async def morts(self, ctx)
        vivants = self.cog.vivants
        morts = self.cog.morts

        # no players
        ctx = mock_discord.get_ctx(vivants)
        await ctx.invoke()
        ctx.send.assert_called_once()

        ctx = mock_discord.get_ctx(morts)
        await ctx.invoke()
        ctx.send.assert_called_once()

        # vivants only
        mock_bdd.add_campsroles()
        bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                   statut=bdd.Statut.vivant).add()
        bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2",
                   chambre="ch2", statut=bdd.Statut.vivant).add()

        ctx = mock_discord.get_ctx(vivants)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertIn("ch2", ctx.send.call_args.args[0])

        ctx = mock_discord.get_ctx(morts)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertNotIn("Joueur1", ctx.send.call_args.args[0])

        # all status
        bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3",
                   statut=bdd.Statut.mort).add()
        bdd.Joueur(discord_id=4, chan_id_=41, nom="Joueur4",
                   chambre="ch4", statut=bdd.Statut.mort).add()
        bdd.Joueur(discord_id=5, chan_id_=51, nom="Joueur5",
                   statut=bdd.Statut.MV).add()
        bdd.Joueur(discord_id=6, chan_id_=61, nom="Joueur6",
                   statut=bdd.Statut.immortel).add()

        ctx = mock_discord.get_ctx(vivants)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur4", ctx.send.call_args.args[0])
        self.assertNotIn("ch4", ctx.send.call_args.args[0])
        self.assertIn("Joueur5", ctx.send.call_args.args[0])
        self.assertIn("Joueur6", ctx.send.call_args.args[0])

        ctx = mock_discord.get_ctx(morts)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Joueur3", ctx.send.call_args.args[0])
        self.assertIn("Joueur4", ctx.send.call_args.args[0])
        self.assertNotIn("ch4", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur5", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur6", ctx.send.call_args.args[0])

        # more players
        for i in range(100):
            bdd.Joueur(discord_id=100+i, chan_id_=i, nom=f"Joueur{100+i}",
                       chambre=f"ch{100+i}", statut=bdd.Statut.vivant).add()
            bdd.Joueur(discord_id=200+i, chan_id_=i, nom=f"Joueur{200+i}",
                       statut=bdd.Statut.mort).add()

        ctx = mock_discord.get_ctx(vivants)
        await ctx.invoke()
        ctx.send.assert_called()
        sent = "\n".join(call.args[0] for call in ctx.send.call_args_list)
        for i in range(100):
            self.assertIn(f"Joueur{100+i}", sent)
            self.assertIn(f"ch{100+i}", sent)

        ctx = mock_discord.get_ctx(morts)
        await ctx.invoke()
        ctx.send.assert_called()
        sent = "\n".join(call.args[0] for call in ctx.send.call_args_list)
        for i in range(100):
            self.assertIn(f"Joueur{200+i}", sent)



import unittest
from unittest import mock

from lgrez.blocs import ready_check



class Test_RCDict(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.ready_check._RCDict class."""

    def test___init__(self):
        """Unit tests for _RCDict.__init__ method."""
        # def __init__(self, _is_ready=None, _class=None, **kwargs)
        __init__ = ready_check._RCDict.__init__

        # simplest
        slf = ready_check._RCDict.__new__(ready_check._RCDict)
        __init__(slf)
        self.assertEqual(slf._errormsg, "No attribute")
        self.assertTrue(slf._is_ready(False))
        self.assertTrue(slf._is_ready(2))
        self.assertTrue(slf._is_ready("a"))
        self.assertFalse(slf._is_ready(None))

        # simple
        slf = ready_check._RCDict.__new__(ready_check._RCDict)
        dic = {str(i): mock.Mock() for i in range(5)}
        __init__(slf, **dic)
        for i in range(5):
            self.assertEqual(slf[str(i)], dic[str(i)])
        self.assertEqual(slf._errormsg, "No attribute")
        self.assertTrue(slf._is_ready(False))
        self.assertTrue(slf._is_ready(2))
        self.assertTrue(slf._is_ready("a"))
        self.assertFalse(slf._is_ready(None))

        # custom _is_ready
        slf = ready_check._RCDict.__new__(ready_check._RCDict)
        dic = {str(i): mock.Mock() for i in range(5)}
        _is_ready = mock.Mock()
        __init__(slf, **dic, _is_ready=_is_ready)
        for i in range(5):
            self.assertEqual(slf[str(i)], dic[str(i)])
        self.assertEqual(slf._errormsg, "No attribute")
        self.assertEqual(slf._is_ready, _is_ready)

        # custom _is_ready and _class
        slf = ready_check._RCDict.__new__(ready_check._RCDict)
        dic = {str(i): mock.Mock() for i in range(5)}
        _is_ready = mock.Mock()
        _class = mock.Mock(__qualname__="zbrrr")
        __init__(slf, **dic, _is_ready=_is_ready, _class=_class)
        for i in range(5):
            self.assertEqual(slf[str(i)], dic[str(i)])
        self.assertIn("zbrrr", slf._errormsg)
        self.assertEqual(slf._is_ready, _is_ready)


    def test__get_raw(self):
        """Unit tests for _RCDict._get_raw method."""
        # def _get_raw(self, name)
        _get_raw = ready_check._RCDict._get_raw
        slf = ready_check._RCDict(a=1, b=2)
        # existing
        val = _get_raw(slf, "a")
        self.assertEqual(val, 1)
        # non-existing
        with self.assertRaises(AttributeError):
            _get_raw(slf, "c")


    def test___getitem__(self):
        """Unit tests for _RCDict.__getitem__ method."""
        # def __getitem__(self, name)
        __getitem__ = ready_check._RCDict.__getitem__
        slf = mock.Mock(ready_check._RCDict, _get_raw=mock.Mock(),
                        _is_ready=mock.Mock())
        # inexistant
        slf._get_raw.side_effect = AttributeError
        with self.assertRaises(AttributeError):
            __getitem__(slf, "c")           # c -> "ccc" < 4
        # ready
        slf._get_raw.side_effect = lambda name: name*3
        slf._is_ready.side_effect = lambda name: len(name) > 4
        val = __getitem__(slf, "ro")
        self.assertEqual(val, "rororo")     # "rororo" > 4
        # not ready
        with self.assertRaises(ready_check.NotReadyError):
            __getitem__(slf, "c")           # c -> "ccc" < 4



class Test_RCMeta(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.ready_check._RCMeta class."""

    def test___new__(self):
        """Unit tests for _RCMeta.__new__ method."""
        # def __new__(metacls, name, bases, dict, check=None, check_type=None)
        __new__ = ready_check._RCMeta.__new__

        # simple
        result = __new__(ready_check._RCMeta, "ah", (), {})
        self.assertTrue(isinstance(result, ready_check._RCMeta))
        self.assertEqual(result.__name__, "ah")

        # complex
        result = __new__(ready_check._RCMeta, "ah", (int,), {"_a": 1, "b": 2},
                         check="fsodfpsd", check_type=dict)
        self.assertTrue(isinstance(result, ready_check._RCMeta))
        self.assertEqual(result.__name__, "ah")
        self.assertEqual(result.__bases__, (int,))
        self.assertIn(("_a", 1), result.__dict__.items())
        self.assertNotIn(("b", 2), result.__dict__.items())   # b public!


    @mock.patch("lgrez.blocs.ready_check._RCDict")
    def test___init__(self, rdc_patch):
        """Unit tests for _RCMeta.__init__ method."""
        # def __init__(cls, name, bases, dict, check=None, check_type=None)
        __new__ = ready_check._RCMeta.__new__
        __init__ = ready_check._RCMeta.__init__

        # simple
        cls = __new__(ready_check._RCMeta, "ah", (), {})
        __init__(cls, "ah", (), {})
        rdc_patch.assert_called_once_with(_is_ready=None, _class=cls)
        rdc_patch.reset_mock()

        # with dict and check
        check = mock.Mock()
        cls = __new__(ready_check._RCMeta, "ah", (int,), {"_a": 1, "b": 2},
                      check=check)
        __init__(cls, "ah", (int,), {"_a": 1, "b": 2}, check=check)
        self.assertIn(("_a", 1), cls.__dict__.items())
        self.assertNotIn(("b", 2), cls.__dict__.items())
        rdc_patch.assert_called_once_with(_is_ready=check, _class=cls, b=2)
                                          # called with public only
        self.assertEqual(cls._rc_dict, rdc_patch.return_value)
        rdc_patch.reset_mock()

        # with dict and check_type
        class Oui:
            pass
        cls = __new__(ready_check._RCMeta, "ah", (int,), {"_a": 1, "b": 2},
                      check_type=Oui)
        __init__(cls, "ah", (int,), {"_a": 1, "b": 2}, check_type=Oui)
        self.assertIn(("_a", 1), cls.__dict__.items())
        self.assertNotIn(("b", 2), cls.__dict__.items())
        rdc_patch.assert_called_once_with(_is_ready=mock.ANY, _class=cls, b=2)
                                          # called with public only
        self.assertEqual(cls._rc_dict, rdc_patch.return_value)
        _is_ready = rdc_patch.call_args.kwargs["_is_ready"]
        self.assertFalse(_is_ready(1))
        self.assertFalse(_is_ready("a"))
        self.assertFalse(_is_ready(None))
        self.assertTrue(_is_ready(Oui()))
        rdc_patch.reset_mock()

        # with dict, check and check_type
        class Oui(str):
            pass
        cls = __new__(ready_check._RCMeta, "ah", (int,), {"_a": 1, "b": 2},
                      check=lambda oui: len(oui) > 3, check_type=Oui)
        __init__(cls, "ah", (int,), {"_a": 1, "b": 2},
                 check=lambda oui: len(oui) > 3, check_type=Oui)
        self.assertIn(("_a", 1), cls.__dict__.items())
        self.assertNotIn(("b", 2), cls.__dict__.items())
        rdc_patch.assert_called_once_with(_is_ready=mock.ANY, _class=cls, b=2)
                                          # called with public only
        self.assertEqual(cls._rc_dict, rdc_patch.return_value)
        _is_ready = rdc_patch.call_args.kwargs["_is_ready"]
        self.assertFalse(_is_ready(1))
        self.assertFalse(_is_ready("a"))
        self.assertFalse(_is_ready("aaaaaaaa"))
        self.assertFalse(_is_ready(None))
        self.assertFalse(_is_ready(Oui()))
        self.assertFalse(_is_ready(Oui("aa")))
        self.assertTrue(_is_ready(Oui("aaaa")))
        rdc_patch.reset_mock()


    def test___getattr__(self):
        """Unit tests for _RCMeta.__getattr__ method."""
        # def __getattr__(cls, name)
        __getattr__ = ready_check._RCMeta.__getattr__
        cls = mock.Mock(_rc_dict={"a": 1, "b": 2})
        # private
        with self.assertRaises(AttributeError):
            __getattr__(cls, "_c")
        # not found
        with self.assertRaises(KeyError):
            __getattr__(cls, "d")
        # found
        result = __getattr__(cls, "a")
        self.assertEqual(result, 1)

    def test___setattr__(self):
        """Unit tests for _RCMeta.__setattr__ method."""
        # def __setattr__(cls, name, value)
        __setattr__ = ready_check._RCMeta.__setattr__
        cls = mock.Mock(_rc_dict={"a": 1, "b": 2})
        # private -> setattr
        with self.assertRaises(TypeError):
            # try to setattr, cant 'cause of mocking
            __setattr__(cls, "_c", "oh")
        # public -> _rc_dict
        result = __setattr__(cls, "a", "eh")
        self.assertNotEqual(cls.a, "eh")
        self.assertEqual(cls._rc_dict["a"], "eh")

    def test___delattr__(self):
        """Unit tests for _RCMeta.__delattr__ method."""
        # def __delattr__(cls, name)
        __delattr__ = ready_check._RCMeta.__delattr__
        cls = mock.Mock(_rc_dict={"a": 1, "b": 2})
        # private -> delattr
        with self.assertRaises(TypeError):
            # try to delattr, cant 'cause of mocking
            __delattr__(cls, "_c")
        # public -> _rc_dict
        __delattr__(cls, "a")
        self.assertNotIn("a", cls._rc_dict)

    def test___iter__(self):
        """Unit tests for _RCMeta.__iter__ method."""
        # def __iter__(cls)
        __iter__ = ready_check._RCMeta.__iter__
        cls = mock.Mock(_rc_dict={"a": 1, "b": 2})
        result = __iter__(cls)
        self.assertEqual(list(result), list(iter(cls._rc_dict)))

    def test_get_raw(self):
        """Unit tests for _RCMeta.get_raw method."""
        # def get_raw(cls, attr)
        get_raw = ready_check._RCMeta.get_raw
        cls = mock.Mock(_rc_dict=mock.Mock(_get_raw=mock.Mock()))
        attr = mock.Mock()
        result = get_raw(cls, attr)
        cls._rc_dict._get_raw.assert_called_once_with(attr)
        self.assertEqual(result, cls._rc_dict._get_raw.return_value)



"""lg-rez / features / Commandes de gestion des salons

Création, ajout, suppression de membres

"""

import asyncio
import functools
import datetime

from discord.ext import commands

from lgrez import config
from lgrez.blocs import tools, one_command
from lgrez.bdd import Joueur, Boudoir, Bouderie
from lgrez.features.sync import transtype



def in_boudoir(callback):
    """Décorateur : commande utilisable dans un boudoir uniquement.

    Lors d'une invocation de la commande décorée hors d'un boudoir
    (enregistré dans :class:`.bdd.Boudoir`), affiche un message d'erreur.

    Ce décorateur n'est utilisable que sur une commande définie dans un Cog.
    """
    @functools.wraps(callback)
    async def new_callback(cog, ctx, *args, **kwargs):
        try:
            Boudoir.from_channel(ctx.channel)
        except ValueError:
            await ctx.reply("Cette commande est invalide en dehors "
                            "d'un boudoir.")
        else:
            return await callback(cog, ctx, *args, **kwargs)

    return new_callback


def gerant_only(callback):
    """Décorateur : commande utilisable par le gérant d'un boudoir uniquement.

    Lors d'une invocation de la commande décorée par un membre qui n'est
    pas gérant du boudoir, affiche un message d'erreur.

    Ce décorateur doittoujours être utilisé en combinaison avec
    :func:`in_boudoir` et positionné après lui.

    Ce décorateur n'est utilisable que sur une commande définie dans un Cog.
    """
    @functools.wraps(callback)
    async def new_callback(cog, ctx, *args, **kwargs):
        boudoir = Boudoir.from_channel(ctx.channel)
        gerant = Joueur.from_member(ctx.author)
        if boudoir.gerant != gerant:
            await ctx.reply("Seul le gérant du boudoir peut utiliser "
                            "cette commande.")
        else:
            return await callback(cog, ctx, *args, **kwargs)

    return new_callback


async def _invite(joueur, boudoir, invite_msg):
    """Invitation d'un joueur dans un boudoir (lancer comme tâche à part)"""
    pc = joueur.private_chan
    bc = boudoir.chan
    mess = await pc.send(
        f"{joueur.member.mention} {boudoir.gerant.nom} t'as invité(e) à "
        f"rejoindre son boudoir : « {boudoir.nom} » !\nAcceptes-tu ?"
    )

    if await tools.yes_no(mess):
        ok = await boudoir.add_joueur(joueur)
        if ok:
            info = f"{joueur.nom} a rejoint le boudoir !"
            confirm = f"Tu as bien rejoint {bc.mention} !"
        else:
            info = None
            confirm = f"Impossible de rejoindre le boudoir."
    else:
        info = f"{joueur.nom} a refusé l'invitation à rejoindre ce boudoir."
        confirm = "Invitation refusée."

    if info:
        try:
            await invite_msg.reply(info)
        except discord.HTTPException:       # Message d'inviation supprimé
            await bc.send(info)

    await mess.reply(confirm)



class GestionChans(commands.Cog):
    """Gestion des salons"""

    @commands.group(aliases=["boudoirs"])
    async def boudoir(self, ctx):
        """Gestion des boudoirs

        Les options relatives à un boudoir précis ne peuvent être
        exécutées que dans ce boudoir ; certaines sont réservées au
        gérant dudit boudoir.
        """
        if not ctx.invoked_subcommand:
            # Pas de sous-commande (correspondante)
            if ctx.subcommand_passed:
                # Tentative de subcommand
                raise commands.BadArgument(
                    f"Option '{ctx.subcommand_passed}' inconnue"
                )
            ctx.message.content = f"!help {ctx.invoked_with}"
            with one_command.bypass(ctx):
                await config.bot.process_commands(ctx.message)


    @boudoir.command(aliases=["liste"])
    @tools.joueurs_only
    @tools.private
    async def list(self, ctx):
        """Liste les boudoirs dans lesquels tu es"""
        joueur = Joueur.from_member(ctx.author)
        bouderies = joueur.bouderies

        if not bouderies:
            await ctx.reply(
                "Tu n'es dans aucun boudoir pour le moment.\n"
                f"{tools.code('!boudoir create')} pour en créer un."
            )
            return

        rep = "Tu es dans les boudoirs suivants :"
        for bouderie in bouderies:
            rep += f"\n - {bouderie.boudoir.chan.mention}"
            if bouderie.gerant:
                rep += " (gérant)"

        rep += "\n\nUtilise `!boudoir leave` dans un boudoir pour le quitter."

        await ctx.send(rep)


    @boudoir.command(aliases=["new", "creer", "créer"])
    @tools.vivants_only
    @tools.private
    async def create(self, ctx, *, nom=None):
        """Crée un nouveau boudoir dont tu es gérant"""
        member = ctx.author
        joueur = Joueur.from_member(member)

        if not nom:
            await ctx.send("Comment veux-tu nommer ton boudoir ?\n"
                           + tools.ital("(`stop` pour annuler)"))
            mess = await tools.wait_for_message_here(ctx)
            nom = mess.content

        if len(nom) > 32:
            await ctx.send("Le nom des boudoirs est limité à 32 caractères.")
            return

        await ctx.send("Création du boudoir...")
        async with ctx.typing():
            now = datetime.datetime.now()
            categ = await tools.multicateg(config.boudoirs_category_name)
            chan = await config.guild.create_text_channel(
                nom,
                topic=f"Boudoir crée le {now:%d/%m à %H:%M}. "
                      f"Gérant(e) : {joueur.nom}",
                category=categ,
            )

            boudoir = Boudoir(chan_id=chan.id, nom=nom, ts_created=now)
            boudoir.add()
            await boudoir.add_joueur(joueur, gerant=True)

            await chan.send(
                f"{member.mention}, voici ton boudoir ! "
                "Tu peux maintenant y inviter des gens avec la commande "
                "`!boudoir invite`."
            )

        await ctx.send(f"Ton boudoir a bien été créé : {chan.mention} !")
        await tools.log(f"Boudoir {chan.mention} créé par {joueur.nom}.")


    @boudoir.command(aliases=["add"])
    @tools.joueurs_only
    @in_boudoir
    @gerant_only
    async def invite(self, ctx, *, cible=None):
        """Invite un joueur à rejoindre ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        joueur = await tools.boucle_query_joueur(
            ctx, cible=cible, message="Qui souhaites-tu inviter ?"
        )
        if joueur in boudoir.joueurs:
            await ctx.send(f"{joueur.nom} est déjà dans ce boudoir !")
            return

        mess = await ctx.send(f"Invitation envoyée à {joueur.nom}.")
        asyncio.create_task(_invite(joueur, boudoir, mess))
        # On envoie l'invitation en arrière-plan (libération du chan).


    @boudoir.command(aliases=["remove", "kick"])
    @tools.joueurs_only
    @in_boudoir
    @gerant_only
    async def expulse(self, ctx, *, cible=None):
        """Expulse un membre de ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        joueur = await tools.boucle_query_joueur(
            ctx, cible=cible, message="Qui souhaites-tu expulser ?"
        )
        if joueur not in boudoir.joueurs:
            await ctx.send(f"{joueur.nom} n'est pas membre du boudoir !")
            return

        await boudoir.remove_joueur(joueur)
        await joueur.private_chan.send(f"Tu as été expulsé(e) du boudoir "
                                       f"« {boudoir.nom} ».")
        await ctx.send(f"{joueur.nom} a bien été expulsé de ce boudoir.")


    @boudoir.command(aliases=["quit"])
    @tools.joueurs_only
    @in_boudoir
    async def leave(self, ctx):
        """Quitte ce boudoir"""
        joueur = Joueur.from_member(ctx.author)
        boudoir = Boudoir.from_channel(ctx.channel)

        if boudoir.gerant == joueur:
            await ctx.send(
                "Tu ne peux pas quitter un boudoir que tu gères. "
                "Utilise `!boudoir transfer` pour passer les droits "
                "de gestion ou `!boudoir delete` pour le supprimer."
            )
            return

        mess = await ctx.reply(
            "Veux-tu vraiment quitter ce boudoir ? Tu ne "
            "pourras pas y retourner sans invitation."
        )
        if not await tools.yes_no(mess):
            await ctx.send("Mission aborted.")
            return

        await boudoir.remove_joueur(joueur)
        await ctx.send(tools.ital(f"{joueur.nom} a quitté ce boudoir."))


    @boudoir.command(aliases=["transmit"])
    @tools.joueurs_only
    @in_boudoir
    @gerant_only
    async def transfer(self, ctx, cible=None):
        """Transfère les droits de gestion de ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        gerant = Joueur.from_member(ctx.author)
        joueur = await tools.boucle_query_joueur(
            ctx, cible=cible, message=("À qui souhaites-tu confier "
                                       "la gestion de ce boudoir ?")
        )
        if joueur not in boudoir.joueurs:
            await ctx.send(f"{joueur.nom} n'est pas membre de ce boudoir !")
            return

        mess = await ctx.reply(
            "Veux-tu vraiment transférer les droits de ce boudoir ? "
            "Tu ne pourras pas les récupérer par toi-même."
        )
        if not await tools.yes_no(mess):
            await ctx.send("Mission aborted.")
            return

        bd_gerant = next(bd for bd in boudoir.bouderies if bd.joueur == gerant)
        bd_nouv = next(bd for bd in boudoir.bouderies if bd.joueur == joueur)

        bd_gerant.gerant = False
        bd_nouv.gerant = True
        bd_nouv.ts_promu = datetime.datetime.now()
        Bouderie.update()
        await boudoir.chan.edit(
            topic=f"Boudoir crée le {boudoir.ts_created:%d/%m à %H:%M}. "
                  f"Gérant(e) : {joueur.nom}"
        )

        await ctx.send(f"Boudoir transféré à {joueur.nom}.")


    @boudoir.command()
    @tools.joueurs_only
    @in_boudoir
    @gerant_only
    async def delete(self, ctx):
        """Supprime ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        mess = await ctx.reply("Veux-tu vraiment supprimer ce boudoir ? "
                               "Cette action est irréversible.")
        if not await tools.yes_no(mess):
            await ctx.send("Mission aborted.")
            return

        await ctx.send("Suppression...")
        for joueur in boudoir.joueurs:
            await boudoir.remove_joueur(joueur)
            await joueur.private_chan.send(
                f"Le boudoir « {boudoir.nom } » a été supprimé."
            )

        await boudoir.chan.edit(name=f"\N{CROSS MARK} {boudoir.nom}")
        await ctx.send(tools.ital(
            "[Tous les joueurs ont été exclus de ce boudoir ; "
            "le channel reste présent pour archive.]"
        ))


    @boudoir.command()
    @tools.joueurs_only
    @in_boudoir
    @gerant_only
    async def rename(self, ctx, *, nom=None):
        """Renomme ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        if not nom:
            await ctx.send("Comment veux-tu renommer ce boudoir ?\n"
                           + tools.ital("(`stop` pour annuler)"))
            mess = await tools.wait_for_message_here(ctx)
            nom = mess.content

        if len(nom) > 32:
            await ctx.send("Le nom des boudoirs est limité à 32 caractères.")
            return

        boudoir.nom = nom
        boudoir.update()
        await boudoir.chan.edit(name=nom)
        await ctx.send("Boudoir renommé avec succès.")


    @boudoir.command(aliases=["hého"])
    @tools.joueurs_only
    @in_boudoir
    @gerant_only
    async def ping(self, ctx, *, mess=""):
        """Mentionne tous les joueurs vivants dans le boudoir."""
        await ctx.channel.send(f"{config.Role.joueur_en_vie.mention} {mess}")


    @commands.command()
    @tools.mjs_only
    async def addhere(self, ctx, *joueurs):
        """Ajoute les membres au chan courant (COMMANDE MJ)

        Args:
            *joueurs: membres à ajouter, chacun entouré par des
                guillemets si nom + prénom

        Si ``*joueurs`` est un seul élément, il peut être de la forme
        ``<crit>=<filtre>`` tel que décrit dans l'aide de ``!send``.
        """
        ts_debut = ctx.message.created_at - datetime.timedelta(microseconds=1)

        if len(joueurs) == 1 and "=" in joueurs[0]:
            # Si critère : on remplace joueurs
            crit, _, filtre = joueurs[0].partition("=")
            crit = crit.strip()
            if crit in Joueur.attrs:
                col = Joueur.attrs[crit]
                arg = transtype(filtre.strip(), col)
                joueurs = Joueur.query.filter_by(**{crit: arg}).all()
            else:
                raise commands.UserInputError(f"critère '{crit}' incorrect")
        else:
            # Sinon, si noms / mentions
            joueurs = [await tools.boucle_query_joueur(ctx, cible)
                       for cible in joueurs]

        for joueur in joueurs:
            await ctx.channel.set_permissions(joueur.member,
                                              read_messages=True)
            await ctx.send(f"{joueur.nom} ajouté")

        mess = await ctx.send("Fini, purge les messages ?")
        if await tools.yes_no(mess):
            await ctx.channel.purge(after=ts_debut)


    @commands.command()
    @tools.mjs_only
    async def purge(self, ctx, N=None):
        """Supprime tous les messages de ce chan (COMMANDE MJ)

        Args:
            N: nombre de messages à supprimer (défaut : tous)
        """
        if N:
            mess = await ctx.send(
                f"Supprimer les {N} messages les plus récents de ce chan ? "
                "(sans compter le `!purge` et ce message)"
            )
        else:
            mess = await ctx.send("Supprimer tous les messages de ce chan ?")

        if await tools.yes_no(mess):
            await ctx.channel.purge(limit=int(N) + 2 if N else None)



"""lg-rez / features / Process d'inscription

Étapes préliminaires pour les joueurs avant le début de saison

"""

import asyncio

from lgrez import config
from lgrez.blocs import tools, env, gsheets
from lgrez.bdd import Joueur, Role, Camp, Statut


async def new_channel(member):
    """Crée et renvoie un nouveau salon privé.

    Peut être étendue, mais toujours appeller cette fonction pour
    créer le chan en lui-même, au risque d'altérer le fonctionnement
    normal du bot.

    Args:
        member (discord.Member): le membre pour qui créer le salon.

    Returns:
        :class:`discord.TextChannel`
    """
    categ = await tools.multicateg(config.private_chan_category_name)
    chan = await member.guild.create_text_channel(
        f"{config.private_chan_prefix}{member.name}",
        topic=str(member.id),       # topic provisoire : ID du membre
        category=categ,
    )
    await chan.set_permissions(member, read_messages=True,
                               send_messages=True)

    return chan


async def register_on_tdb(joueur):
    """Enregistre un joueur dans le Tableau de bord.

    Peut être personnalisé à un autre système de gestion des joueurs.

    Args:
        joueur (.bdd.Joueur): le joueur à enregistrer.

    Note:
        Fonction asynchrone depuis la version 2.2.2.
    """
    SHEET_ID = env.load("LGREZ_TDB_SHEET_ID")
    workbook = await gsheets.connect(SHEET_ID)
    sheet = await workbook.worksheet(config.tdb_main_sheet)
    values = await sheet.get_all_values()           # Liste de listes

    head = values[config.tdb_header_row - 1]
    # Ligne d'en-têtes (noms des colonnes), - 1 car indexé à 0

    id_index = gsheets.a_to_index(config.tdb_id_column)
    pk = head[id_index]
    if pk != Joueur.primary_col.key:
        raise ValueError(
            "Tableau de bord : la cellule "
            "`config.tdb_id_column` / `config.tdb_header_row` = "
            f"`{config.tdb_id_column}{config.tdb_header_row}` "
            f"vaut `{pk}` au lieu de la clé primaire de la table "
            f"`Joueur`, `{Joueur.primary_col.key}` !"
        )

    mstart, mstop = config.tdb_main_columns
    main_indexes = range(gsheets.a_to_index(mstart),
                         gsheets.a_to_index(mstop) + 1)
    # Indices des colonnes à remplir
    cols = {}
    for index in main_indexes:
        col = head[index]
        if col in Joueur.attrs:
            cols[col] = Joueur.attrs[col]
        else:
            raise ValueError(
                f"Tableau de bord : l'index de la zone principale "
                f"`{col}` n'est pas une colonne de la table `Joueur` !"
                " (voir `lgrez.config.main_indexes` / "
                "`lgrez.config.tdb_header_row`)"
            )

    tstart, tstop = config.tdb_tampon_columns
    tampon_indexes = range(gsheets.a_to_index(tstart),
                           gsheets.a_to_index(tstop) + 1)

    TDB_tampon_index = {}
    for index in tampon_indexes:
        col = head[index].partition("_")[2]
        if col in cols:
            TDB_tampon_index[col] = index
        else:
            raise ValueError(
                f"Tableau de bord : l'index de zone tampon `{head[index]}` "
                f"réfère à la colonne `{col}` (partie suivant le premier "
                f"underscore), qui n'est pas une colonne de la zone "
                "principale ! (voir `lgrez.config.tampon_indexes` / "
                "`lgrez.config.main_indexes`)"
            )

    plv = config.tdb_header_row         # Première Ligne Vide
    # (si tableau vide, suit directement le header, - 1 pour l'indexage)
    for i_row, row in enumerate(values):
        # On parcourt les lignes du TDB
        if i_row < config.tdb_header_row:
            # Ligne avant le header / le header (car décalage de 1)
            continue

        if row[id_index].isdigit():
            # Si il y a un vrai ID dans la colonne ID
            plv = i_row + 1

    modifs = [gsheets.Modif(plv, id_index, joueur.discord_id)]
    for index in main_indexes:
        val = getattr(joueur, head[index])
        modifs.append(gsheets.Modif(plv, index, val))
    for index in tampon_indexes:
        val = getattr(joueur, head[index].partition("_")[2])
        # Colonnes "tampon_<col>" ==> <col>
        modifs.append(gsheets.Modif(plv, index, val))

    try:
        await gsheets.update(sheet, *modifs)
    except gsheets.ConnectionError:
        await joueur.private_chan.send("Petit problème, on réessaie...")
        await asyncio.sleep(10)
        await gsheets.update(sheet, *modifs)


# Routine d'inscription (fonction appellée par la commande !co)
async def main(member):
    """Routine d'inscription complète d'un joueur.

    Args:
        member (:class:`~discord.Member`): joueur à inscrire.

    Crée et paramètre le salon privé du joueur, lui pose des questions
    et l'inscrit en base.

    Personalisation : voir :obj:`.config.demande_chambre`,
    :obj:`.config.chambre_mj`, :func:`.config.additional_inscription_step`
    et :obj:`.config.debut_saison`.

    Commande appellée à l'arrivée sur le serveur, utiliser
    :meth:`\!co <.bot.Special.Special.co.callback>` pour trigger cette
    commande depuis Discord.
    """
    try:
        joueur = Joueur.from_member(member)
    except ValueError:      # Joueur pas encore inscrit en base
        pass
    else:                   # Joueur dans la bdd = déjà inscrit
        chan = joueur.private_chan
        await chan.set_permissions(member, read_messages=True,
                                   send_messages=True)
        await chan.send(
            f"Saloww ! {member.mention} tu es déjà inscrit, "
            "viens un peu ici !"
        )
        return

    if (chan := tools.get(config.guild.text_channels, topic=str(member.id))):
        # Inscription en cours (topic du chan = ID du membre)
        await chan.set_permissions(member, read_messages=True,
                                   send_messages=True)
        await chan.send(
            f"Tu as déjà un channel à ton nom, {member.mention}, par ici !"
        )

    else:           # Pas d'inscription déjà en cours : création channel
        chan = await new_channel(member)

    # Récupération nom et renommages

    await chan.send(
        f"Bienvenue {member.mention} ! Je suis le bot à qui tu auras "
        "affaire tout au long de la partie.\nTu es ici sur ton channel "
        "privé, auquel seul toi et les MJ ont accès ! C'est ici que tu "
        "lanceras toutes les commandes pour voter, te renseigner... mais "
        "aussi que les MJ discuteront avec toi en cas de soucis.\n\nPas "
        "de panique, je vais tout t'expliquer !"
    )

    await tools.sleep(chan, 5)

    await chan.send(
        "Avant toute chose, finalisons ton inscription !\nD'abord, un "
        "point règles nécessaire :\n\n"
        + tools.quote_bloc(
            "En t'inscrivant au Loup-Garou de la Rez, tu garantis vouloir "
            "participer à cette édition et t'engages à respecter les "
            "règles du jeu. Si tu venais à entraver le bon déroulement de "
            "la partie pour une raison ou une autre, les MJ "
            "s'autorisent à te mute ou t'expulser du Discord sans préavis."
        )
    )

    await tools.sleep(chan, 5)

    message = await chan.send(
        "C'est bon pour toi ?\n"
        + tools.ital(
            "(Le bot te demandera souvent confirmation, en t'affichant "
            "deux réactions comme ceci. Clique sur ✅ si ça te va, sur "
            "❎ sinon. Tu peux aussi répondre (oui, non, ...) par écrit.)"
        )
    )
    if not await tools.yes_no(message):
        await chan.send(
            "Pas de soucis. Si tu changes d'avis ou que c'est un "
            f"missclick, appelle un MJ aled ({tools.code('@MJ')})."
        )
        return

    await chan.send(
        "Parfait. Je vais d'abord avoir besoin de ton (vrai) prénom, "
        "celui par lequel on t'appelle au quotidien. Attention, tout "
        "troll sera foudracanonné™ !"
    )

    def check_chan(m):
        # Message de l'utilisateur dans son channel perso
        return m.channel == chan and m.author != config.bot.user

    ok = False
    while not ok:
        await chan.send(
            "Quel est ton prénom, donc ?\n"
            + tools.ital("(Répond simplement dans ce channel, "
                         "à l'aide du champ de texte normal)")
        )
        prenom = await tools.wait_for_message(check=check_chan, chan=chan)

        await chan.send("Très bien, et ton nom de famille ?")
        nom_famille = await tools.wait_for_message(check=check_chan, chan=chan)
        nom = f"{prenom.content.title()} {nom_famille.content.title()}"
        # .title met en majuscule la permière lettre de chaque mot

        message = await chan.send(
            f"Tu me dis donc t'appeller {tools.bold(nom)}. "
            "C'est bon pour toi ? Pas d'erreur, pas de troll ?"
        )
        ok = await tools.yes_no(message)

    await chan.edit(name=config.private_chan_prefix + nom)  # Renommage conv
    if member.top_role < config.Role.mj:
        # Renommage joueur (ne peut pas renommer les MJ)
        await member.edit(nick=nom)

    await chan.send(
        "Parfait ! Je t'ai renommé(e) pour que tout le monde te "
        "reconnaisse, et j'ai renommé cette conversation."
    )

    if config.demande_chambre:
        await tools.sleep(chan, 3)
        a_la_rez = await tools.yes_no(await chan.send(
            "Bien, dernière chose : habites-tu à la Rez ?"
        ))

        if a_la_rez:
            def sortie_num_rez(m):
                # Longueur de chambre de rez maximale
                return len(m.content) < Joueur.chambre.type.length
            mess = await tools.boucle_message(
                chan, "Alors, quelle est ta chambre ?", sortie_num_rez,
                rep_message=("Désolé, ce n'est pas un numéro de chambre "
                             "valide, réessaie...")
            )
            chambre = mess.content
        else:
            chambre = config.chambre_mj

    else:
        chambre = None

    res = await config.additional_inscription_step(member, chan)
    if res is False:
        return

    if config.demande_chambre:
        await chan.send(
            f"{nom}, en chambre {chambre}... Je t'inscris en base !"
        )
    else:
        await chan.send(f"{nom}... Je t'inscris en base !")

    # Indicateur d'écriture pour informer le joueur que le bot fait des trucs
    async with chan.typing():
        # Enregistrement en base

        joueur = Joueur(
            discord_id=member.id,
            chan_id_=chan.id,
            nom=member.display_name,
            chambre=chambre,
            statut=Statut.vivant,
            role=Role.default(),
            camp=Camp.default(),
            votant_village=True,
            votant_loups=False,
            role_actif=False,
        )
        joueur.add()

        # Ajout sur le TDB

        await register_on_tdb(joueur)

        # Grant accès aux channels joueurs et information

        await member.add_roles(config.Role.joueur_en_vie)
        await chan.edit(topic="Ta conversation privée avec le bot, "
                              "c'est ici que tout se passera !")

    # Conseiller d'ACTIVER TOUTES LES NOTIFS du chan
    # (et mentions only pour le reste, en activant @everyone)
    await chan.send(
        f"Tu es maintenant inscrit(e) ! Je t'ai attribué le rôle "
        f"{config.Role.joueur_en_vie.mention}, qui te donne l'accès à "
        "tout plein de nouveaux channels à découvrir."
    )
    await chan.send(
        "Juste quelques dernières choses :\n "
        "- Plein de commandes te sont d'ores et déjà accessibles ! "
        f"Découvre les toutes en tapant {tools.code('!help')} ;\n "
        "- Si tu as besoin d'aide, mentionne "
        f"simplement les MJs ({tools.code('@' + config.Role.mj.name)}) "
        "et on viendra voir ce qui se passe !\n "
        "- Si ce n'est pas le cas, je te conseille fortement d'installer "
        "Discord sur ton téléphone, "
        f"et d'{tools.bold('activer toutes les notifications')} pour ce "
        "channel ! Promis, pas de spam :innocent:\n"
        "Pour le reste du serveur, tu peux le mettre en \"mentions only\", "
        f"en activant le {tools.code('@everyone')} – il est limité ;\n\n"
        "Enfin, n'hésite pas à me parler, j'ai toujours quelques réponses "
        "en stock..."
    )

    await tools.sleep(chan, 5)
    await chan.send(
        "Voilà, c'est tout bon ! Installe toi bien confortablement, "
        f"la partie commence le {config.debut_saison}."
    )

    # Log
    await tools.log(
        f"Inscription de {member.name}#{member.discriminator} "
        f"réussie\n - Nom : {nom}\n - Chambre : {chambre}\n"
        f" - Channel créé : {chan.mention}"
    )



"""lg-rez / blocs / Limitation à une commande

Système maison pour ne pas pouvoir utiliser plus d'une commande
à la fois

"""

import asyncio

from discord.ext import commands

from lgrez import config


#: list[~discord.ext.command.Command]: Commands exempted from
#: one_command limitation, registered by :func:`.do_not_limit`
exempted = []


class AlreadyInCommand(commands.CheckFailure):
    """Salon déjà occupé par une commande non finissable.

    Exception levée lorsqu'un membre veut lancer une commande dans un
    salon où une commande est déjà en cours d'exécution, et que cette
    commande n'a pas pu être arrêtée (ou pas assez rapidement) par
    un message ``"stop"``.

    Dérive de :exc:`discord.ext.commands.CheckFailure`.
    """
    pass


async def not_in_command(ctx):
    """Check : assure qu'une commande n'est pas en cours dans ce salon.

    Fonction à utiliser comme check global, pour toutes les commandes
    (enregistrer avec :meth:`~discord.ext.commands.Bot.add_check`)

    Args:
        ctx (discord.ext.commands.Context): contexte d'invocation
            de la commande.

    Returns:
        ``True``

    Raises:
        :exc:`AlreadyInCommand`.
    """
    if ctx.command in exempted:
        return True         # Commandes exemptées
    if ctx.channel.id not in config.bot.in_command:
        return True         # Channel libre

    # On envoie (discrètement) l'ordre d'arrêter la commande précédente
    await ctx.send(config.stop_keywords[0], delete_after=0)
    # On attend qu'il soit pris en compte
    await asyncio.sleep(1)

    if ctx.channel.id in config.bot.in_command:    # Si ça n'a pas suffit
        raise AlreadyInCommand()                    # on raise l'erreur

    return True


# @bot.before_invoke
async def add_to_in_command(ctx):
    """Ajoute le channel à la liste des channels dans une commande.

    Fonction à appeller avant chaque appel de fonction
    (enregistrer avec :meth:`~discord.ext.commands.Bot.before_invoke`)

    Elle est appellée seulement si les checks sont OK, donc pas si le
    salon est déjà dans :attr:`config.bot.in_command <.LGBot.in_command>`.

    Args:
        ctx (discord.ext.commands.Context): contexte d'invocation de
            la commande.
    """
    if ctx.command not in exempted and not ctx.message.webhook_id:
        config.bot.in_command.append(ctx.channel.id)


# @bot.after_invoke
async def remove_from_in_command(ctx):
    """Retire le channel de la liste des channels dans une commande.

    Fonction à appeller après chaque appel de fonction.
    (enregistrer avec :meth:`~discord.ext.commands.Bot.after_invoke`)

    Elle attend 0.1 secondes avant d'enlever le joueur afin d'éviter
    que le bot réagisse « nativement » (IA) à un message déjà traité
    par un :func:`.tools.wait_for_message` ayant mené à la fin de la
    commande.

    Args:
        ctx (discord.ext.commands.Context): contexte d'invocation de
            la commande.
    """
    await asyncio.sleep(0.1)        # On attend un peu
    if (ctx.channel.id in config.bot.in_command
        and ctx.command not in exempted):

        config.bot.in_command.remove(ctx.channel.id)


class _Bypasser():
    def __init__(self, ctx):
        self.ctx = ctx

    def __enter__(self):
        config.bot.in_command.remove(self.ctx.channel.id)
        return self

    def __exit__(self, exc_type, exc, tb):
        config.bot.in_command.append(self.ctx.channel.id)


def bypass(ctx):
    """Context manager: bypass one-command limitation.

    Args:
        ctx (discord.ext.commands.Context): running command context.

    Use in a command callback to launch a second one without problems::

        with one_command.bypass(ctx):
            await config.bot.process_commands(some_message)

    """
    return _Bypasser(ctx)


def do_not_limit(command):
    """Decorator for commands not concerned by one_command limitation.

    Register the command in :attr:`.one_command.exempted`.

    Args:
        command (discord.ext.commands.Command): the command to register.

    Returns:
        :class:`discord.ext.commands.Command`
    """
    if not isinstance(command, commands.Command):
        raise TypeError("one_command.not_limited is a decorator for Commands")

    exempted.append(command)
    return command



"""lg-rez / features / Tâches planifiées

Planification, liste, annulation, exécution de tâches planifiées

"""

import datetime

from discord.ext import commands

from lgrez.blocs import tools
from lgrez.bdd import Tache, Action


class GestionTaches(commands.Cog):
    """Commandes de planification, exécution, annulation de tâches"""

    @commands.command()
    @tools.mjs_only
    async def taches(self, ctx):
        """Liste les tâches en attente (COMMANDE MJ)

        Affiche les commandes en attente d'exécution (dans la table
        :class:`.bdd.Tache`) et le timestamp d'exécution associé.
        Lorsque la tâche est liée à une action, affiche le nom de
        l'action et du joueur concerné.
        """
        async with ctx.typing():
            lst = Tache.query.order_by(Tache.timestamp).all()
            LT = ""
            for tache in lst:
                LT += (
                    f"\n{str(tache.id).ljust(5)} "
                    f"{tache.timestamp.strftime('%d/%m/%Y %H:%M:%S')}    "
                    f"{tache.commande.ljust(25)} ")
                if (action := tache.action):
                    LT += f"{action.base.slug.ljust(20)} {action.joueur.nom}"

            if LT:
                mess = (
                    "Tâches en attente : \n\nID    Timestamp              "
                    "Commande                  Action               Joueur"
                    f"\n{'-'*105}{LT}\n\n"
                    "Utilisez !cancel <ID> pour annuler une tâche."
                )
            else:
                mess = "Aucune tâche en attente."

        await tools.send_code_blocs(ctx, mess)


    @commands.command(aliases=["doat"])
    @tools.mjs_only
    async def planif(self, ctx, quand, *, commande):
        """Planifie une tâche au moment voulu (COMMANDE MJ)

        Args:
            quand: format ``[<J>/<M>[/<AAAA>]-]<H>:<M>[:<S>]``,
                avec ``<J>``, ``<M>``, ``<AAAA>``, ``<H>`` et ``<M>``
                des entiers et ``<S>`` un entier/flottant optionnel.
                La date est optionnelle (défaut : date du jour).
                Si elle est précisée, elle doit être **séparée de
                l'heure par un tiret** et l'année peut être omise ;
            commande: commande à exécuter (commençant par un ``!``).
                La commande sera exécutée PAR UN WEBHOOK dans LE CHAN
                ``#logs`` : toutes les commandes qui sont liées au
                joueur ou réservées au chan privé sont à proscrire
                (ou doivent a minima être précédées de ``!doas cible``)

        Cette commande repose sur l'architecture en base de données,
        ce qui garantit l'exécution de la tâche même si le bot plante
        entre temps.

        Si le bot est down à l'heure d'exécution prévue, la commande
        sera exécutée dès le bot de retour en ligne.

        Si la date est dans le passé, la commande est exécutée
        immédiatement.

        Examples:
            - ``!planif 18:00 !close maire``
            - ``!planif 13/06-10:00 !open maire``
            - ``!planif 13/06/2020-10:00 !open maire``
            - ``!planif 23:25:12 !close maire``
        """
        now = datetime.datetime.now()

        if "/" in quand:            # Date précisée
            date, time = quand.split("-")
            J, MA = date.split("/", maxsplit=1)
            day = int(J)
            if "/" in MA:           # Année précisée
                M, A = MA.split("/")
                month = int(M)
                year = int(A)
            else:
                month = int(MA)
                year = now.year
            date = datetime.date(year=year, month=month, day=day)
        else:
            date = now.date()
            time = quand

        H, MS = time.split(":", maxsplit=1)
        hour = int(H)
        if ":" in MS:               # Secondes précisées
            M, S = MS.split(":")
            minute = int(M)
            second = int(S)
        else:
            minute = int(MS)
            second = 0
        time = datetime.time(hour=hour, minute=minute, second=second)

        ts = datetime.datetime.combine(date, time)

        action_id = None
        # ID de l'action associée à la tâche le cas échéant
        try:
            quoi, id = commande.split()
            if quoi in ["!open", "!close", "!remind"]:
                action_id = int(id)
        except ValueError:
            pass

        if ts < datetime.datetime.now():
            mess = await ctx.send(
                "Date dans le passé ==> exécution immédiate ! On valide ?"
            )
            if not await tools.yes_no(mess):
                await ctx.send("Mission aborted.")
                return

        tache = Tache(timestamp=ts,
                      commande=commande,
                      action=Action.query.get(action_id))
        tache.add()         # Planifie la tâche
        await ctx.send(
            f"{tools.code(commande)} planifiée pour le "
            f"{tools.code(ts.strftime('%d/%m/%Y %H:%M:%S'))}.\n"
            f"{tools.code(f'!cancel {tache.id}')} pour annuler."
        )


    @commands.command(aliases=["retard", "doin"])
    @tools.mjs_only
    async def delay(self, ctx, duree, *, commande):
        """Exécute une commande après XhYmZs (COMMANDE MJ)

        Args:
            quand: format ``[<X>h][<Y>m][<Z>s]``, avec ``<X>`` (heures)
                et ``<Y>`` (minutes) des entiers et ``<Z>`` (secondes)
                un entier ou un flottant. Chacune des trois composantes
                est optionnelle, mais au moins une doit être présente ;
            commande: commande à exécuter (commençant par un ``!``).
                La commande sera exécutée PAR UN WEBHOOK dans LE CHAN
                ``#logs`` : toutes les commandes qui sont liées au
                joueur ou réservées au chan privé sont à proscrire
                (ou doivent être précédées d'un ``!doas <cible>``).

        Cette commande repose sur l'architecture en base de données,
        ce qui garantit l'exécution de la commande même si le bot plante
        entre temps.

        Si le bot est down à l'heure d'exécution prévue, la commande
        sera exécutée dès le bot de retour en ligne.

        Examples:
            - ``!delay 2h !close maire``
            - ``!delay 1h30m !doas @moi !vote Yacine Oussar``
        """
        secondes = 0
        try:
            if "h" in duree.lower():
                h, duree = duree.split("h")
                secondes += 3600*int(h)
            if "m" in duree.lower():
                m, duree = duree.split("m")
                secondes += 60*int(m)
            if "s" in duree.lower():
                s, duree = duree.split("s")
                secondes += float(s)
        except Exception as e:
            raise commands.BadArgument("<duree>") from e

        if duree or not secondes:
            raise commands.BadArgument("<duree>")

        ts = datetime.datetime.now() + datetime.timedelta(seconds=secondes)
        action_id = None
        # ID de l'action associée à la tâche le cas échéant
        try:
            quoi, id = commande.split(" ")
            if quoi in ["!open", "!close", "!remind"]:
                action_id = int(id)
        except ValueError:
            pass

        tache = Tache(timestamp=ts,
                      commande=commande,
                      action=Action.query.get(action_id))
        tache.add()         # Planifie la tâche

        await ctx.send(
            f"Commande {tools.code(commande)} planifiée pour le "
            f"{tools.code(ts.strftime('%d/%m/%Y %H:%M:%S'))}\n"
            f"{tools.code(f'!cancel {tache.id}')} pour annuler."
        )


    @commands.command()
    @tools.mjs_only
    async def cancel(self, ctx, *ids):
        """Annule une ou plusieurs tâche(s) planifiée(s) (COMMANDE MJ)

        Args:
            *ids: IDs des tâches à annuler, séparées par des espaces.

        Utiliser ``!taches`` pour voir la liste des IDs.
        """
        taches = []
        for id in ids:
            if id.isdigit() and (tache := Tache.query.get(int(id))):
                taches.append(tache)

        if not taches:
            await ctx.send("Aucune tâche trouvée.")
            return

        message = await ctx.send("Annuler les tâches :\n" + "\n".join([
            f" - {tools.code(tache.timestamp.strftime('%d/%m/%Y %H:%M:%S'))} "
            f"> {tools.code(tache.commande)}" for tache in taches
        ]))
        if not await tools.yes_no(message):
            await ctx.send("Mission aborted.")
            return

        Tache.delete(*taches)       # Annule les tâches
        await ctx.send("Tâche(s) annulée(s).")



"""lg-rez / blocs / Console d'administration (avancé)

Permet de contrôler le bot depuis la console le faisant tourner.

"""

import sys
import asyncio

import aioconsole
import asyncode
import discord

from lgrez import config


class CaptureLogs():
    """Context manager: capture the messages posted to a Discord channel.

    Only capture messages posted by the bot itself: these messages
    are not sent at all to Discord.

    Args:
        chan (discord.TextChannel): the channel to capture.
    """
    def __init__(self):
        """Initialize self."""
        self._chan = config.Channel.logs
        self._calls = []

        async def _cor(text, *args, **kwargs):
            self._calls.append(text)

        class _ProxyChannel(discord.TextChannel):
            def __init__(self):
                # We inherit TextChannel for type checks, but we
                # do NOT want to create a new TextChannel object
                pass
            def __getattribute__(slf, attr):
                if attr == "send":
                    return _cor
                return getattr(self._chan, attr)

        self.proxy = _ProxyChannel()

    def __enter__(self):
        """Enter the context: start capturing."""
        config.Channel.logs = self.proxy
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context: stop capturing."""
        config.Channel.logs = self._chan

    @property
    def text(self):
        """str: The messages sent in the channel, joined with newlines."""
        return "\n".join(self._calls)


class AdminConsole(asyncode.AsyncInteractiveConsole):
    """Python console to control LGBot from the running shell.

    (subclass of :class:`asyncode.AsyncInteractiveConsole`)

    Args:
        locals, filename: see :class:`asyncode.AsyncInteractiveConsole`.
    """
    def __init__(self, locals={}, filename="__console__"):
        """Initialize self"""
        locals["_console"] = self
        super().__init__(locals, filename)

    async def interact(self, banner=None, exitmsg="Byebye!"):
        """Launchs the console.

        (see :meth:`asyncode.AsyncInteractiveConsole.interact`)
        """
        await aioconsole.aprint("\nLaunching interactive console...")

        if banner is None:
            base = ('Type "help", "copyright", "credits" or "license" for '
                    'more information.\nasync REPL: use "await" directly '
                    'instead of asyncio.run().')
            warn = ('WARNING: Sending KeyboardInterrupt (^C) will stop '
                    'the whole LGBot.')
            banner = f"Python {sys.version} on {sys.platform}\n{base}\n{warn}"

        try:
            await super().interact(banner, exitmsg)
        except (SystemExit, KeyboardInterrupt, EOFError) as exc:
            yn = await aioconsole.ainput("Exit LG-Rez? (y/N) ")
            if yn.lower().strip() in ("o", "y", "oui", "yes"):
                raise KeyboardInterrupt from None

    async def write(self, data):
        """Method called on each print / repr / traceback...

        Relies on :meth:`aioconsole.aprint`.

        Args:
            data (str): text to display.
        """
        if not hasattr(sys, "ps1"):
            sys.ps1 = '>>> '
        if not hasattr(sys, "ps2"):
            sys.ps2 = '... '

        await aioconsole.aprint(data, end="")

    async def raw_input(self, prompt=""):
        """Method called when the console waits for next instruction.

        Relies on :meth:`aioconsole.ainput`.

        Args:
            prompt (str): optional text to send before waiting for
                input.

        Returns:
            :class:`str`: The new instruction to compute.
        """
        if prompt:
            await self.write(prompt)

        entree = await aioconsole.ainput()

        if entree.startswith(config.bot.command_prefix):
            # Exécution de commande
            entree = await execute_command(entree)

        return entree


async def run_admin_console(locals):
    """Launch the Admin console and restart it if crashing.

    Args:
        locals (dict): the objects accessible from the console.
    """
    while True:
        await asyncio.sleep(1)
        cons = AdminConsole(locals=locals)
        await cons.interact()
        await aioconsole.aprint("Aborted, restarting admin console...")


async def execute_command(entree):
    """Execute a command.

    Create a fake message in the #logs channel by the server owner,
    and process it while capturing the output in the channel.

    Args:
        entree (str): The command to execute (must begin with the
            command prefix)

    Returns:
        str: The messages get in response to the command execution,
        joined with newlines.
    """
    message = (await config.Channel.logs.history(limit=1).flatten())[0]
    # On a besoin de récupérer un message, ici le dernier de #logs
    message.author = config.guild.owner
    message.content = entree
    ctx = await config.bot.get_context(message)

    _calls = []

    async def _cor(text, *args, **kwargs):
        _calls.append(text)

    ctx.send = _cor
    ctx.reply = _cor

    with CaptureLogs() as cc:
        await config.bot.invoke(ctx)

    text = "\n".join(_calls)
    if cc.text:
        text += f"\n[LOGGED] > {cc.text}"
    text = text.replace('"""', '\\"\\"\\"')
    return f'print("""{text}""")'
